#!/cs/local/bin/perl

# the main learning file for CG
# copyright Mirela Andronescu andrones@cs.ubc.ca

# Started more changes on Jan 18, 2008, to make CG work with other programs. I'm trying to make it work with Hosna's Hfold.
# Changes:
#   removed all the default values, they should be taken from the options file. Nothing should be set here, only in the options file.
#   removed initial parameter set option and the variable $initial, instead added a link to the file of initial parameters
#   thermo file has to be given in the options file
#   Apr 7, 2008: added the possibility that train_file is none
# Several other changes were done and not recorded here, see Mirela Andronescu PhD thesis for more details of CG
# Version 2.0 was 

# It is possible to add more iterations to what you ran before, by giving it the same parameters, and increasing the number of iterations

                       
use strict;

# setting default parameters (see below what they mean)

my $min_length_sum = 2000;  # it may also be given in the configuration file
# this is the minimum length sum of the splits, in case we use SGE for prediction and analysis.
# if the predictor is slower, set to a smaller value

my $normalize = "n2";
my $a_param = 2;

my $do_perturbation = 0;     # for now, only works if we use mpi # maybe to take out
my $use_several_computers = 1;
my $simple_model = 0;
my $test_file = "none";

my $fixparams_file = "none";
# the file which contains which parameters are fixed to which values
# by default is none.
my $fixparams_filename = "none";

my $additional_constraints_file = "none";
my $additional_constraints_filename = "none";
my $additional_constraints_lb_file = "none";
my $additional_constraints_ub_file = "none";

my $currdir = `pwd`; chomp ($currdir);
# the current directory


# before I was using MPI to run the prediction and analysis at each step, but that has several disadvantages:
# 1 - it often crashes/hangs
# 2 - it doesn't use sge: unfair to other people, and doesn't load balance well if some machine is already loaded
# now I'm submitting SGE jobs instead.
# I don't think the mpi version works any more, so I wouldn't recommend setting this to 0
my $use_sge_instead_of_mpi = 1;

# flag whether I'm using the arrow cluster for cplex or not.
# I should use it all the time
my $cluster_for_cplex = "icics";  # arrow or icics (includes beta)
my $cluster_for_prediction = "icics";  # arrow or icics (includes beta)
my $there_are_non_zero_means = 0;
my $write_structural_constraints_in_parallel = 0;

# the quadratic/linear program solver. Matlab is another option, not sure it works right now.
# CPLEX is supposed to be much faster.
my $solver = "CPLEX";
my $ipopt_cpp;  # used if solver is IPOPT
my $ipopt_hpp;  # used if solver is IPOPT


# other global variables
my $directory;
my $train_file;
my $train_filename;
my $thermo_XML_file;
my $thermo_XML_filename;
my $thermo_constraints_file;    # the file I will obtain, with the thermodynamic constraints
my $thermo_bounds_file;
my $thermo_precision_file;
my @thermo_precisions = ();
my @thermo_bounds = ();     # for the case when solver eq IPOPT
my $thermo_executable;
my $use_published_errors = 0;
my $num_iterations;
my $amount_lub;
my $lambda;
my $num_train = 0;
my $validation_file = "none";
my $count_threshold;
my $fix_when_counts_0;
my $configuration_file;
my $initial_parameters;
my $executable_constraints;
my $executable_predict_and_analyse;
my $num_split_files = 0;
my $num_split_files_validation = 0;
my $add_regularizer = 0;
my $regularizer_mean_file;
my $regularizer_mean_filename;
my $regularizer_bound_eta;
my $normalize_all_three_sums;
my $scaler;
my $model_options_file = "none";
my $model_options_filename = "none";
my $Xia_weight = 1;
my $Xia_last_exp = 99;
my $max_margin = 0;     # 0 = CG, 1 = DIM-CG, 2 = LAM-CG

# variables related to $max_margin == 1 (DIM-CG)
my $dimcg_exponent = 1;    # 1 = sum of delta; 2 = sum of delta^2 * sign(delta)

# variables related to $max_margin == 2 (LAM-CG)
my $one_delta_permol = 1;   # 0: Use one delta for each structural constraint   # 1: Use one delta per entry, i.e. the same delta for all the constraints corresponding to each entry
my $add_constraint_if_violation = 0;
my $epsilon_for_violation = 0;
my $do_lossaug = 0;
my $use_equalities = 0;

# The usage message
# input can be: an options file
if ($#ARGV != 0)
{
    print "Usage: $0 <options_file>\n";
    #print "<directory> is the directory where all the files refered to from the options file are located\n";
    print "<options_file> is the configuration file\n";
    exit;
}


# read the input options file, has to be provided
$configuration_file = $ARGV[0];
# assume the configuration file is $directory/config/file. So now I'm recovering $directory.
if ($configuration_file =~ /^(\S+)\/config\//)
{
    $directory = $1;
}
else
{
    print "The configuration file should be directory/config/file.\n";
    exit;
}

# read all the options provided in the configuration file
my $line;
open (FILE, $configuration_file) || die ("Cannot open $configuration_file");
while ($line = <FILE>)
{
    if ($line =~ /^#/)
    {   next; }
    #if ($line =~ /SGE cluster: (\S+)\n/)    # I used to have this, but I changed it "Cluster for prediction"
    #{   $which_cluster = $1;    }
    elsif ($line =~ /Training file: (.*)\n/)
    {
        if ($train_file ne "none")
        {
            $train_file = $1;
            $num_train = get_num_points("$directory/$train_file");
        }
        print "Number of training points: $num_train\n";
    }
    elsif ($line =~ /Minimum length sum: (\d+)/)
    {   $min_length_sum = $1;   }
    elsif ($line =~ /Training filename: (\S*)/)
    {   $train_filename = $1;   }
    elsif ($line =~ /Number of iterations.*: (\S*)/)
    {   $num_iterations = $1;   }
    elsif ($line =~ /Bounds parameter B.*: (\S*)/)
    {   $amount_lub = $1;   }
    elsif ($line =~ /Weight of the thermodynamic set lambda.*: (\S*)/)
    {   $lambda = $1;   }
    elsif ($line =~ /Add regularizer: (\d+)/)
    {   $add_regularizer = $1; }
    elsif ($line =~ /Regularizer mean file: (\S+)/)
    {   $regularizer_mean_file = $1;   }
    elsif ($line =~ /Regularizer mean filename: (\S+)/)
    {   $regularizer_mean_filename = $1;   }
    elsif ($line =~ /Regularizer bound eta: (\S+)/)
    {   $regularizer_bound_eta = $1; }
    elsif ($line =~ /Validation file: (\S*)/)
    {   $validation_file = $1;   }
    elsif ($line =~ /Test file: (\S*)/)
    {   $test_file = $1;   }
    elsif ($line =~ /Initial params file: (\S*)/)
    {   $initial_parameters = $1;  }
    elsif ($line =~ /Fixparams file: (\S*)/)
    {   $fixparams_file = $1;  }
    elsif ($line =~ /Fixparams filename: (\S*)/)
    {   $fixparams_filename = $1;  }
    elsif ($line =~ /Additional constraints file: (\S*)/)
    {   $additional_constraints_file = $1;  }
    # If the constraints do not = 0, then a lower bound file and an upper bound file must be given.
    elsif ($line =~ /Additional constraints filename: (\S*)/)
    {   $additional_constraints_filename = $1;  }
    elsif ($line =~ /Additional constraints lower bounds file: (\S*)/)
    {   $additional_constraints_lb_file = $1;  }
    elsif ($line =~ /Additional constraints upper bounds file: (\S*)/)
    {   $additional_constraints_ub_file = $1;  }
    elsif ($line =~ /Thermo XML file: (\S*)/)
    {   $thermo_XML_file = $1;  }
    elsif ($line =~ /Thermo XML filename: (\S*)/)
    {   $thermo_XML_filename = $1;  }
    elsif ($line =~ /Executable that creates the thermodynamic constraints: (\S*)/)
    {   $thermo_executable = $1;  }
    elsif ($line =~ /Use published errors: (\d)/)
    {   $use_published_errors = $1; }    
    elsif ($line =~ /L norm used for the objective function.*: (\S*)/)
    {   $a_param = $1;   }
    elsif ($line =~ /Normalisation criterion.*: (\S*)/)
    {   $normalize = $1;   }
    elsif ($line =~ /Model is simplified.*: (\S*)/)
    {   $simple_model = $1;   }
    elsif ($line =~ /Do perturbation.*: (\S*)/)
    {   $do_perturbation = $1;   }
    elsif ($line =~ /Executable that creates the structural constraints: (\S+)/)
    {   $executable_constraints = $1;   }
    elsif ($line =~ /Executable that predicts and analyses results: (\S+)/)
    {   $executable_predict_and_analyse = $1;   }
    elsif ($line =~ /Normalize all three sums: (\d)/)
    {   $normalize_all_three_sums = $1;     }
    elsif ($line =~ /Cluster for cplex: (\S+)/)
    {   $cluster_for_cplex = $1;    }
    elsif ($line =~ /Cluster for prediction: (\S+)/)
    {   $cluster_for_prediction = $1;    }
    #Write structural constraints in parallel: 1
    elsif ($line =~ /Write structural constraints in parallel: (\d)/)
    {   $write_structural_constraints_in_parallel = $1;    }
    elsif ($line =~ /Solver: (\S+)/)
    {   $solver = $1;   }    
    elsif ($line =~ /Model options file: (\S+)/)
    {   $model_options_file = $1;   }
    elsif ($line =~ /Model options filename: (\S+)/)
    {   $model_options_filename = $1;   }
    elsif ($line =~ /Xia_Turner_1998 weight: (\d+)/)
    {   $Xia_weight = $1;   }
    #Use a max margin approach: 
    elsif ($line =~ /Use a max margin approach: (\d)/)
    {   $max_margin = $1;   }
    elsif ($line =~ /DIM-CG exponent: (\d)/)
    {   $dimcg_exponent = $1;  }
    elsif ($line =~ /Use one delta per molecule: (\d)/)
    {   $one_delta_permol = $1;     }
    elsif ($line =~ /Add constraint only if new constraint violates condition: (\d)/)
    {   $add_constraint_if_violation = $1;  }
    # 0: don't check if it violates condition
    # 1: check, following Taskar's Phd thesis, page 52.
    # For now, it's only applied to LAM-CG    
    elsif ($line =~ /Epsilon for violation: (\d)/)
    {   $epsilon_for_violation = $1;    }
    # Epsilon is a positive number or 0
    # For now, it's only applied to LAM-CG
    elsif ($line =~ /Use equality constraints instead of inequality constraints: (\d)/)
    {   $use_equalities = $1;   }

    
    #elsif ($line =~ /Dangling ends are fixed to the initial values.*: (.*)\n/)
    #{   $fix_dangling_ends = $1;   }
}
close (FILE);

if ($normalize_all_three_sums)      {   $scaler = 1000; }
else                                {   $scaler = 1;    }

if ($max_margin == 2)   {   $do_lossaug = 1;    }

####################################

# other options that I tried in the past and that are fixed for now. I am not sure if changing them works any more.

my $fix_when_counts_0 = 0;
#     print "\t<fix_when_counts_0:0|1> 1 means that if a feature doesn't appear in the thermodynamic set nor\n";
#     print "\t\tin the structural set, it is fixed to the value of the initial parameter set\n";
#     print "\t\t0 means all those values are free. DEFAULT: $fix_when_counts_0\n";

#my $validation_file = "none";
#     print "\t<validation_file> is a file in the same format as <train_file>. If both sensitivity and \n";
#     print "\t\tppv are higher than for the initial parameters, then we compute them for this file too.\n";
#     print "\t\t<validation_file> can be none if no such computation is desired.\n";

my $count_threshold = 0;
#     print "\t<count_threshold> was 0 for the ISMB paper\n";

# try to replace the constraints instead of adding them (doesn't give good results at all)
# the option was <replace_instead_of_add:{1|0}> it should always be add
my $replace_instead_of_add = 0;
  
# if the bounds B are from the initial set of parameters (T) or from the previous set (P)  
my $bounds_from_T_or_P = "T";
# fix bounds: fixb-yes means fixed bounds across iterations
# fixb-yes: fixed bounds, the ones specified. NO means they decrease by .5k
my $fixb = "fixb-yes";    
# other bounds stuff
my $iter_same_bound = 10;    # how many iterations to perform with the same bound
my $bounds_relax = 0.5;      # by how much to enlarge/relax the bounds

# tried a regularizer for the parameters, didn't really work    DEPRICATED
#my $regularizer_beta = 0;

# I tried to normalize the least squares equation (objective function) by using a length dependence
# lendep-yes: means add length dependence in the objective function
my $lendep = "lendep-no"; 

# OTHER OPTIONS TO TRY, MAYBE

# try adding only those constraints which are different than the true structure at least by some threshold
# already implemented in create_structural_constraints

# for the features which do not appear at all, should we fix them to the initial parameters? Or should we let them free?

####################################

# other global variables
my $continue = 0;
my $problem_file;
my $problem_file_nopath;
my $answer_file;
my $answer_file_nopath;
my $basis_file;     # this was a good idea to use basis files, for better CPLEX speed, but only for linear programs (i.e. a=1)
my $prev_basis_file;
my $optimizer;
my $newdir;
my $num_thermo_constr = 0;
my $num_struct_constr = 0;
my $num_additional_constr = 0;
my $opt_problem;
my $line;
my $strlen;
#my $test_set;
my $num_params;
my $num_zero_params = 0;    # if some variables are fixed to 0 (such as dangling ends), this variable gets incremented. Used to normalize the ridge, so that it doesn't depend on the number of parameters

############################################

# validate some of the input options

if ($cluster_for_cplex ne "arrow" && $cluster_for_cplex ne "icics")
{
    print "Cluster for cplex must be arrow or icics, ABORT!\n";
    exit;
}
if ($cluster_for_prediction ne "arrow" && $cluster_for_prediction ne "icics")
{
    print "Cluster for prediction must be arrow or icics, ABORT!\n";
    exit;
}
if ($amount_lub !~ /[%K]/ && $amount_lub ne "inf" && $amount_lub ne "custom" && $amount_lub ne "custom2")
{
    print "<amount_lub> must be e.g. 0.8%, 0.8K, inf, custom or custom2\n";
    exit;
}
if ($bounds_from_T_or_P ne "T" && $bounds_from_T_or_P ne "P")
{
    print "bounds_from_T_or_P must be T or P\n";
    exit;
}

if ($solver ne "CPLEX" and $solver ne "IPOPT")
{
    print "solver must be IPOPT or CPLEX. MATLAB is not supported any more!\n";
    exit;
}
if ($normalize ne "n1" and $normalize ne "n2")
{
    print "normalizer must be n1 (1/D) or n2 (1/D*1/si)\n";
    exit;
}
if ($a_param != 1 and $a_param != 2)
{
    print "a_param must be 1 or 2\n";
    exit;
}

if ($lendep ne "lendep-yes" and $lendep ne "lendep-no")
{
    print "lendep must be yes or no\n";
    exit;
}
if ($fixb ne "fixb-yes" and $fixb ne "fixb-no")
{
    print "fixb must be yes or no\n";
    exit;
}

############################################

# prepare some more variables

# set the $opt_problem to lp or qp
if ($a_param == 1)
{
    $opt_problem = "lp";    
}
else
{
    $opt_problem = "qp";    
}    

# not used any more
# my $fm_or_sm;
# if ($simple_model)
# {    $fm_or_sm = "sm";    }
# else
# {    $fm_or_sm = "fm";    }

$optimizer = "optimize";
# -${test_set}-${num_test} missing, but we don't need it anyway

# the new output directory name
# create a complicated name, that includes most of the input options
$newdir = "$currdir/$directory/results/RES_${train_filename}";
my $maybe_dir = "$currdir/$directory/results/RES_${train_filename}";
# by adding $currdir, we make sure this is an absolute dir name, to prevent problems when the directories are moved

$newdir .= "_it${num_iterations}_bnd${amount_lub}_${normalize}_a${a_param}";
$maybe_dir .= "_it*_bnd${amount_lub}_${normalize}_a${a_param}";

if ($model_options_file ne "none")
{
    $newdir .= "_$model_options_filename";
    $maybe_dir .= "_$model_options_filename";
}
if ($lambda != 0)
{
    $newdir .= "_$thermo_XML_filename";
    $maybe_dir .= "_$thermo_XML_filename";
    if ($Xia_weight > 1)
    {
        $newdir .= "_Xia${Xia_weight}";
        $maybe_dir .= "_Xia${Xia_weight}";
    }
}
$newdir .= "_lambda${lambda}_p${use_published_errors}";
$maybe_dir .= "_lambda${lambda}_p${use_published_errors}";
if ($add_regularizer == 1 || $add_regularizer == 2)     # 3 is not implemented yet
{
    $newdir .=    "_REG-L${add_regularizer}-mean${regularizer_mean_filename}-eta${regularizer_bound_eta}";
    $maybe_dir .= "_REG-L${add_regularizer}-mean${regularizer_mean_filename}-eta${regularizer_bound_eta}";
}
if ($fixparams_file ne "none")
{
    $newdir .= "_${fixparams_filename}";
    $maybe_dir .= "_${fixparams_filename}";
}
if ($additional_constraints_file ne "none")
{
    $newdir .= "_${additional_constraints_filename}";
    $maybe_dir .= "_${additional_constraints_filename}";    
}

$newdir .= "_N${normalize_all_three_sums}";
$maybe_dir .= "_N${normalize_all_three_sums}";

if ($max_margin)
{
    $newdir .= "_MM${max_margin}";
    $maybe_dir .= "_MM${max_margin}";
    if ($max_margin == 1)
    {
        if ($dimcg_exponent == 2)
        {
            $newdir .= "-sq";
            $maybe_dir .= "-sq";
        }
    }
    if ($max_margin == 2)
    {
        $newdir .= "_WITH-LOSS-SCALED_";       # added on Sep 3, 2008. I had forgotten to add the loss in the cnstraint before!!!
        # _WITH-LOSS_ is when I add -distance to the rhs of the constraints
        # _WITH-LOSS-SCALED_ is when I add -distance/length to the rhs of the constraints
        if ($use_equalities)     {    $newdir .= "-eq";   $maybe_dir .= "-eq";   }
        else                     {    $newdir .= "-ineq";   $maybe_dir .= "-ineq";   }
        if ($one_delta_permol)   {    $newdir .= "-d-one";   $maybe_dir .= "-d-one";   }
        else                     {    $newdir .= "-d-many";  $maybe_dir .= "-d-many";   }
        $newdir .= "-add${add_constraint_if_violation}-eps${epsilon_for_violation}";
        $maybe_dir .= "-add${add_constraint_if_violation}-eps${epsilon_for_violation}";
    }
}

my $iteration = 1;
my $command;
my $datafile; 
my $new_datafile;
my $datadir; 
my $new_datadir;
my $params_file;
my $new_params_file;
my $results_filename = "results_final.txt";
my $results_file = "$newdir/$results_filename";
my $counts_file;    # not used any more
my $line_lb;
my $line_ub;
my $init_indent = "  ";
my $indent = "";
my $truestr = "";
my $index = 0;
my $num_unmet_params = 0;
my $num_constraints_permol_filename;
my $num_molecules;
my @num_constraints_permol;
my $sge_predict_and_analyse;
my $sge_validate;   # essentially the same as $sge_predict_and_analyse, except for the case when we do loss-augmented prediction, we only want it for $sge_predict_and_analyse.
my $sge_constraints;


my $thermo_obj;
if ($a_param == 2)
{ $thermo_obj = "l2norm"; }
else
{ $thermo_obj = "l1norm"; }

# check to see if the dir we want already exists - means we already ran some iterations, or even all
if (-e $newdir)
{
    $continue = 1;
    print "Directory already exists, continue to add more suboptimals\n";
}
# or check to see if we ran the same thing, but with different number of iterations
else
{
    # check if we already ran a few iterations
    my @lls = `ls $maybe_dir 2>/dev/null`;
    if ($#lls > -1)
    {
        $continue = 1;
        print "Directory already exists, continue to add more suboptimals\n";
        if ($maybe_dir ne $newdir)
        {
            system ("mv $maybe_dir $newdir");
        }        
    }
}

if ( ! $continue)
{
    $command = "mkdir -p $newdir";
    print "$command\n";
    system ($command);
    system ("cp start_sge_job.pl $newdir");

    # copy the configuration file, to know which one I used
    system ("cp $configuration_file $newdir");    

    #system ("rm -rf results");
    #system ("ln -s $newdir results");
    `echo >$results_file`;
    $command = "cp $directory/$initial_parameters $newdir/params_sub-0.txt";
    print "$command\n";
    system($command);

    if ($train_file ne "none")
    {
        $command = "cp $directory/$train_file $newdir/set0.txt";
        print "$command\n";
        system($command);
    }
      
    # if we want to use sge, also create a directory with many files, one per sequence/structure
    if ($use_sge_instead_of_mpi && $train_file ne "none")    # use sge
    {
        $command = "perl split_dataset_into_many_files.pl $min_length_sum $directory/$train_file $newdir/setdir0";
        print "$command\n";
        $num_split_files = `$command`;
        chomp ($num_split_files);
    }      

    if ($validation_file ne "none")
    {
        $command = "cp $directory/$validation_file $newdir/validation_set.txt";
        print "$command\n";
        system($command);
        # also split the file
        if ($use_sge_instead_of_mpi)
        {
            $command = "perl split_dataset_into_many_files.pl $min_length_sum $newdir/validation_set.txt $newdir/valdir0";
            print "$command\n";
            $num_split_files_validation = `$command`;
            chomp ($num_split_files_validation);
            # also create the tmp directory
            system ("mkdir $newdir/valdir0/tmp");
        }
    }

    if ($test_file ne "none")
    {
        $command = "cp $directory/$test_file $newdir/test_set.txt";
        print "$command\n";
        system($command);
    }
}

####################################

# write down the cluster name in a file, so that I can change it during run time
open (CLUSTER, ">$newdir/cluster.txt") || die ("Cannot open file $newdir/cluster.txt");
print CLUSTER "$cluster_for_prediction";
close (CLUSTER);
# NOTE: I haven't tested this carefully, not sure if it works in case you do change it at runtime

####################################

# find out $num_params
$num_params = `cat $directory/$initial_parameters | wc -l`;
chomp ($num_params);
print "Num params = $num_params\n";

####################################

# an output log file, so that you can track what's going on
my $output_verbose = "output_verbose.txt";
my $options_file = "$newdir/$output_verbose";

if (!$continue)
{
    # write all the options in a file, so that I know exactly what options I used.    
    open (OPT, ">$options_file") || die ("Cannot open $options_file");
    print OPT "INPUT OPTIONS\n";
    print OPT "Training file: $train_file\n";
    print OPT "Training filename: $train_filename\n";
    print OPT "Number of iterations: $num_iterations\n";
    print OPT "Bounds parameter B: $amount_lub\n";
    print OPT "Weight of the thermodynamic set lambda: $lambda\n";
    print OPT "Use the published errors for the thermo set: $use_published_errors\n";
    print OPT "Add regularizer: $add_regularizer\n";
    print OPT "Regularizer mean file: $regularizer_mean_file\n";
    print OPT "Regularizer bound eta: $regularizer_bound_eta\n";
    print OPT "Initial params file: $initial_parameters\n";
    print OPT "Test file: $test_file\n";    
    print OPT "Fixparams file: $fixparams_file\n";
    print OPT "Fixparams filename: $fixparams_filename\n";
    print OPT "Additional constraints file: $additional_constraints_file\n";
    print OPT "Additional constraints filename: $additional_constraints_filename\n";
    print OPT "Additional constraints lower bounds file: $additional_constraints_lb_file\n";
    print OPT "Additional constraints upper bounds file: $additional_constraints_ub_file\n";
    print OPT "Thermo XML file: $thermo_XML_file\n";
    print OPT "Xia_Turner_1998 weight: $Xia_weight\n";
    print OPT "Executable that creates the thermodynamic constraints: $thermo_executable\n";
    print OPT "Number of training points: $num_train\n";
    print OPT "Validation file: $validation_file\n";
    print OPT "Model options file: $model_options_file\n";
    print OPT "Use a max margin approach: $max_margin\n";
    if ($max_margin == 2)
    {
        print OPT "Use one delta per molecule: $one_delta_permol\n";
        print OPT "Add constraint only if new constraint violates condition: $add_constraint_if_violation\n";
        print OPT "Epsilon for violation: $epsilon_for_violation\n";
    }

    print OPT "\nOPTIONAL INPUT OPTIONS\n";
    print OPT "L norm used for the objective function: $a_param\n";
    print OPT "Normalisation criterion: $normalize\n";
    print OPT "Model is simplified: $simple_model\n";    
    print OPT "Count threshold: $count_threshold\n";
    print OPT "Do perturbation: $do_perturbation\n";
    #print OPT "Dangling ends are fixed to the initial values: $fix_dangling_ends\n";
    print OPT "Fix the parameters with counts 0: $fix_when_counts_0\n";
    print OPT "\nOTHER VALUES\n";
    print OPT "Number of parameters in the model: $num_params\n";
    my $date = `date`;
    my $name = `echo \$HOSTNAME`;
    print OPT "Date and time it ran: $date";
    print OPT "Computer name where the job was started: $name";
    if ($use_sge_instead_of_mpi)
    { print OPT "Used SGE for prediction, not MPI\n"; }
    else
    { print OPT "Used MPI for prediction, not SGE\n"; }
    print OPT "Cluster used for cplex: $cluster_for_cplex\n";
    print OPT "Cluster used for prediction: $cluster_for_prediction\n";
    print OPT "Write structural constraints in parallel: $write_structural_constraints_in_parallel\n";
    print OPT "Solver used: $solver\n";
    print OPT "\n\n";
    print OPT "NOW THE FUN BEGINS\n\n\n";
    
    # In case we use mpi, also save the hostfile, so that I know on how many and which computers I ran this
    if (!$use_sge_instead_of_mpi)
    {
        system ("cp hostfile $newdir/hostfile");
    }

    if ($solver eq "IPOPT")
    {
        system ("cp $directory/data/qcp_template.hpp $newdir/qcp.hpp");
        system ("cp $directory/data/qcp_template.opt $newdir/qcp.opt");
        system ("cp $directory/data/qcp_main_template.cpp $newdir/qcp_main.cpp");
        system ("cp $directory/data/Makefile_qcp_template $newdir/Makefile_qcp");
    }
}
else
{
    my $options_file = "$newdir/output_verbose.txt";
    open (OPT, ">>$options_file") || die ("Cannot open $options_file");
    print OPT "\n\n=============================\n\nRUN WAS INTERRUPTED, STARTING AGAIN\n\n";
    print OPT "Number of iterations: $num_iterations\n";
    my $name = `echo \$HOSTNAME`;
    print OPT "Computer name where the job was started: $name";
    print OPT "\n\n";
    print OPT "NOW THE FUN CONTINUES\n\n\n";
}

####################################

if ($train_file ne "none")
{

    # create the sge file in the $newdir directory
    $sge_predict_and_analyse = "sge_predict_and_analyse_results.sh";
    open (SGE, ">$newdir/$sge_predict_and_analyse") || die ("Cannot open $newdir/$sge_predict_and_analyse");
    print SGE "#!/bin/sh\n\n";
    print SGE "# first argument \$1 is the new parameter set\n";
    print SGE "# second argument \$2 is the old directory\n";
    print SGE "# third argument \$3 is the new directory\n";
    print SGE "../../$executable_predict_and_analyse \$1 ";
    if ($model_options_file ne "none")
    {
        print SGE " ../$model_options_file ../../$thermo_XML_file ";
    }
    if ($solver ne "IPOPT")
    {
        print SGE " \$2/file\${SGE_TASK_ID}.txt \$3/file\${SGE_TASK_ID}.txt \$3/res\${SGE_TASK_ID}.txt ";
        if ($do_lossaug)
        {
            print SGE " 1 \$2/delta\${SGE_TASK_ID}.txt ";
        }
    }
    else
    {
        print SGE " \$2/file\${SGE_TASK_ID}.txt \$3/file\${SGE_TASK_ID}.txt \$3/res\${SGE_TASK_ID}.txt ";
    }
    print SGE " >> $results_filename\n\n\n";
    close (SGE);

    $sge_validate = "sge_validate.sh";
    open (SGE, ">$newdir/$sge_validate") || die ("Cannot open $newdir/$sge_validate");
    print SGE "#!/bin/sh\n\n";
    print SGE "# first argument \$1 is the new parameter set\n";
    print SGE "# second argument \$2 is the old directory\n";
    print SGE "# third argument \$3 is the new directory\n";
    print SGE "../../$executable_predict_and_analyse \$1 ";
    if ($model_options_file ne "none")
    {
        print SGE " ../$model_options_file  ../../$thermo_XML_file ";
    }
    print SGE " \$2/file\${SGE_TASK_ID}.txt \$3/file\${SGE_TASK_ID}.txt \$3/res\${SGE_TASK_ID}.txt ";
    print SGE " >> $results_filename 2>&1\n\n\n";
    close (SGE);    
        
    ####################################
        
    # if some iterations happened already, continue adding more
    
    # first, we need to check at which iteration we've left.
    my @all_sets = glob ("$newdir/set*.txt");
    my @numbers = ();
    foreach my $set (@all_sets)
    {
        if ($set =~ /set(\d+)\.txt/)
        {
            push (@numbers, $1+0);
        }    
    }
    #print "@all_sets";
    my @numbers_sorted = sort { $a <=> $b } @numbers;
    $iteration = $numbers_sorted[$#numbers_sorted];
    if (! -e "$newdir/params_sub-$iteration.txt")
    {
        print OPT "ERROR: Parameter file $newdir/params_sub-$iteration.txt doesn't exist\n";
        die "ERROR: Parameter file $newdir/params_sub-$iteration.txt doesn't exist\n";
    }
    $iteration++;
}

# this option is 1 only when the prediction program is HotKnots. 
# The reason is that HotKnots has memory leaks, and it can't write the constraints for more than a couple of input structures, because it runs out of memory. Therefore I'm running it in parallel.
if ($write_structural_constraints_in_parallel)
{
    # create the sge file in the $newdir directory
    $sge_constraints = "sge_constraints.sh";
    open (SGE, ">$newdir/$sge_constraints") || die ("Cannot open $newdir/$sge_constraints");
    print SGE "#!/bin/sh\n\n";
    print SGE "# first argument \$1 is the params file\n";
    print SGE "# first argument \$2 is the input (data) directory\n";
    print SGE "# second argument \$3 is the results directory (with constraints and number of constraints per mol)\n";
    # $command = "$directory/$executable_constraints $params_file $datafile $newdir/tmp $num_constraints_permol_filename";
    print SGE "../../$executable_constraints \$1 ";
    if ($model_options_file ne "none")
    {
        print SGE " $directory/$model_options_file ";
    }
    print SGE " \$2/file\${SGE_TASK_ID}.txt \$3/constr\${SGE_TASK_ID}.txt \$3/constr_ub\${SGE_TASK_ID}.txt \$3/num\${SGE_TASK_ID}.txt >> $results_filename 2>&1\n\n\n";
    close (SGE);
       
}

# create the constraints file
if ($lambda > 0)
{
    $thermo_constraints_file = "$newdir/thermo_${thermo_XML_filename}.txt";
    $thermo_bounds_file = "$newdir/thermo_bounds_${thermo_XML_filename}.txt";           # for IPOPT
    $thermo_precision_file = "$newdir/thermo_precision_${thermo_XML_filename}.txt";
    if ($solver eq "IPOPT")
    {
        $command = "$directory/$thermo_executable $directory/$thermo_XML_file $directory/$initial_parameters $thermo_constraints_file $thermo_bounds_file $thermo_precision_file > /dev/null";
    }
    else
    {
        $command = "$directory/$thermo_executable $directory/$thermo_XML_file ";
        if ($model_options_file ne "none")
        {
            $command .= " $directory/$model_options_file $directory/$initial_parameters ";
        }        
        $command .= " $thermo_constraints_file $thermo_precision_file";
    }
    print "$command\n";
    print OPT "Executing: $command\n";
    system ($command);

    # fill up the @thermo_precisions
    open (PREC, $thermo_precision_file) || die ("Cannot open $thermo_precision_file");
    my $pline;
    my $exp_no = 1;
    while ($pline = <PREC>)
    {
        chomp ($pline);
        if ($use_published_errors)
        {
            if ($exp_no <= $Xia_last_exp)
            {
                push (@thermo_precisions, $pline * $Xia_weight);
            }
            else
            {
                push (@thermo_precisions, $pline);
            }
        }
        else
        {
            if ($exp_no <= $Xia_last_exp)
            {
                push (@thermo_precisions, $Xia_weight);
            }
            else
            {
                push (@thermo_precisions, 1);
            }
        }
        $exp_no++;
    }
    close (PREC);

    # fill up the @thermo_bounds
    if ($solver eq "IPOPT")
    {
        open (FILE, $thermo_bounds_file) || die ("Cannot open $thermo_bounds_file");
        while ($pline = <FILE>)
        {
            chomp ($pline);
            push (@thermo_bounds, $pline);
        }
        close (FILE);
    }    
}

####################################

# figure out what $num_split_files is in the case $continue
if ($continue)
{
    #$command = "ls $newdir/setdir" .($iteration-1). "/file*.txt | wc -l";
    # ls might say too many files
    $command = "cd $newdir/setdir" .($iteration-1). "; find . -name 'file*.txt' | wc -l";
    print "$command\n";
    $num_split_files = `$command`;
    chomp ($num_split_files);

    if ($validation_file ne "none")
    {
        #$command = "ls $newdir/valdir0/file*.txt | wc -l";
        $command = "cd $newdir/valdir0; find . -name 'file*.txt' | wc -l";
        print "$command\n";
        $num_split_files_validation = `$command`;
        chomp ($num_split_files_validation);
    }
}

####################################

# get all_thermodynamic_constraints data
if ($lambda > 0)
{
    if ($solver eq "IPOPT")     # constraints don't have names in this case, but there's one constraint per line
    {
        $num_thermo_constr = `cat $thermo_constraints_file | wc -l`;
    }
    else
    {
        $num_thermo_constr = `grep tc_ $thermo_constraints_file | wc -l`;
    }
    chomp ($num_thermo_constr);
}

if ($additional_constraints_file ne "none")
{
    if ($solver eq "IPOPT")     # constraints don't have equality or inequality signs, but there's one constraint per line
    {
        $num_additional_constr = `cat $directory/$additional_constraints_file | wc -l`;
    }
    else
    {
        $num_additional_constr = `cat $directory/$additional_constraints_file | grep "[<>=]" | wc -l`;
    }
    chomp ($num_additional_constr);
}

# create upper and lower bounds for the energy parameters
if ($amount_lub ne "inf")
{
    $command = "perl get_percentage_lb_ub.pl $directory/$train_file $directory/$initial_parameters $amount_lub $newdir/params_lb_${amount_lub}.txt $newdir/params_ub_${amount_lub}.txt";
    print OPT "$indent$command\n";
    system($command);
}

####################
# now put these bounds in arrays
my @params_lb = ();     # lower bounds
my @params_ub = ();
open (LB, "$newdir/params_lb_${amount_lub}.txt") || die ("Cannot open LB");
open (UB, "$newdir/params_ub_${amount_lub}.txt") || die ("Cannot open UB");
while ($line = <LB>)
{
    chomp ($line);
    push (@params_lb, $line);
    $line = <UB>;
    chomp ($line);
    push (@params_ub, $line);
}
close (LB);
close (UB);

# sanity check
if (@params_lb."" != $num_params || @params_ub."" != $num_params)
{
    print "Number of parameters in params_lb or params_ub, coming from file  $newdir/params_lb_${amount_lub}.txt or $newdir/params_ub_${amount_lub}.txt, it not $num_params! ABORT!\n";
    print OPT "Number of parameters in params_lb or params_ub, coming from file  $newdir/params_lb_${amount_lub}.txt or $newdir/params_ub_${amount_lub}.txt, it not $num_params! ABORT!\n";
    exit;
}
# I have the bounds
#################


# read the fixparams file, if given
my @fixparams = ();
if ($fixparams_file ne "none")
{
    my $fixline;
    open (FIX, "$directory/$fixparams_file") || die ("Cannot open $directory/$fixparams_file");
    while ($fixline = <FIX>)
    {
        chomp ($fixline);
        if ($fixline ne "variable" && $fixline == 0.0)  {   $num_zero_params++;     }
        push (@fixparams, $fixline);
    }    
    close (FIX);
}

print "Num zero params: $num_zero_params\n";
####################################
# now go through each iteration


for (my $s = $iteration; $s <= $num_iterations; $s++)
{       
    # remove the old files
    system ("rm -f $problem_file $num_constraints_permol_filename");
    if ($solver ne "IPOPT")     {   system ("rm -f $answer_file");  }
    # don't remove the problem file for now.
    #system ("rm -f $answer_file $num_constraints_permol_filename");
    $indent = $indent.$init_indent;
    my $tmp =  $s-1;	
    $datafile = "$newdir/set$tmp.txt";
    $new_datafile = "$newdir/set${s}.txt";
    $datadir = "$newdir/setdir$tmp";
    $new_datadir = "$newdir/setdir${s}";
    $params_file = "$newdir/params_sub-${tmp}.txt";
    $new_params_file = "$newdir/params_sub-${s}.txt";
    $counts_file = "$newdir/counts_sub-${s}.txt";
    $num_constraints_permol_filename = "$newdir/numconst_sub-${s}.txt";
    @num_constraints_permol = ();
    my $unique_index = 0;
    my $struct_constr_file = "$newdir/struct_constr.txt";
    # needed for IPOPT
    my $struct_constr_ub_file = "$newdir/struct_constr_ub.txt";
    my $index;
    # read the cluster from the file, just in case it was changed
    if (-e "$newdir/cluster.txt")
    {
        $cluster_for_prediction = `cat $newdir/cluster.txt`;
    }
    my @exists_delta = ();
    

    system ("rm -f $struct_constr_file");
    
    if ($train_file ne "none")
    {
        # presumably the code to generate constraints is in $executable_constraints
        if (!$write_structural_constraints_in_parallel)
        {
            $command = "$directory/$executable_constraints $params_file ";
            if ($model_options_file ne "none")
            {
                $command .= " $directory/$model_options_file ";
            }
            $command .= " $datafile ";

            if ($model_options_file ne "none")
            {
                $command .= " $thermo_XML_file ";
            }
            $command .= " $struct_constr_file $num_constraints_permol_filename";

            if ($max_margin == 2 && $one_delta_permol)
            {
                $command .= " 1";
            }
            print OPT "$indent$command\n";
            my $output = `$command`;

            if ($max_margin == 2 && $one_delta_permol)
            {
                # figure out $exists_delta  (0 if it doesn't, 1 if it exists)
                for (my $i=0; $i < $num_molecules; $i++)
                {
                    $exists_delta[$i] = 0;
                }
                open (CON, $struct_constr_file) || die ("Cannot open $struct_constr_file");
                
                while (my $cline = <CON>)
                {
                    if ($cline =~ /sc(\d+)/)
                    {
                        $exists_delta[$1] = 1;
                    }
                }
                close (CON);
            }
            
            if ($output =~ /ERROR/)
            {
                print OPT "ERROR IN create_structural_constraints, STOP!\n";
                print OPT "$output\n";        
                print "ERROR IN create_structural_constraints, STOP!\n";
                print "$output\n";
                exit;
            }
        }
        else    # do write the constraints in parallel
        # in case we are working with memory-leak predictor, such as HotKnots
        {
            # NEW: in this case, I assume the solver is IPOPT, so then add_structural_constraints outputs a bunch of other things for IPOPT
            if ($solver ne "IPOPT")
            {
                print OPT "ERROR! If write_structural_constraints_in_parallel is 1, the solver must be IPOPT! ABORT!\n";
                print "ERROR! If write_structural_constraints_in_parallel is 1, the solver must be IPOPT! ABORT!\n";
                exit;
            }
            # 1. make directory where I put the files to be split
            # clean up the mess of the previous iteration
            system ("rm -rf $newdir/constrdir_res");
            system ("rm -rf $newdir/constrdir");
            system ("mkdir $newdir/constrdir");
            system ("mkdir $newdir/constrdir_res");

            # 2. split $datafile in many files            
            $command = "perl split_dataset_into_many_files.pl $min_length_sum $datafile $newdir/constrdir";
            #print "$command\n";
            my $num_split_files_constr = `$command`;
            chomp ($num_split_files_constr);

            # prepare the sge command
            if ($cluster_for_prediction eq "icics")
            {
                $command = "cd $newdir; perl start_sge_job.pl 'qsub -cwd -S /bin/sh -o /dev/null -e /dev/null -t 1000-". (1000+$num_split_files_constr-1)." $sge_constraints params_sub-${tmp}.txt constrdir constrdir_res'";
                #$command = "cd $newdir; perl start_sge_job.pl 'qsub -cwd -S /bin/sh -o /dev/null -e /dev/null -t 1000-". (1000+$num_split_files_constr-1)." $sge_constraints params_sub-${tmp}.txt constrdir constrdir_res'";
            }
            elsif ($cluster_for_prediction eq "arrow")
            {
                $command = "nohup ssh samos \"source /cs/beta/lib/pkg/sge-6.0u7_1/default/common/settings.csh; cd $newdir; perl start_sge_job.pl 'qsub -cwd -S /bin/sh -o /dev/null -e /dev/null -p -1 -P kpm -t 1000-". (1000+$num_split_files_constr-1)." $sge_constraints params_sub-${tmp}.txt constrdir constrdir_res'\"";
            }          

            # run the sge job
            print OPT "$indent$command\n";
            system ($command);

            system ("rm -rf $num_constraints_permol_filename $struct_constr_file $struct_constr_ub_file");
            # collect the data
            for (my $i=1000; $i <= 1000+$num_split_files_constr -1; $i++)
            {
                # sanity check: make sure we have the right thing in these files
                # the # rows in constr$i.txt should be the same as the # of rows in constr_ub$i.txt
                #   and should be the same as the sum of values in num$i.txt
                my $num1 = `cat $newdir/constrdir_res/constr$i.txt | wc -l`;
                chomp ($num1);
                my $num2 = `cat $newdir/constrdir_res/constr_ub$i.txt | wc -l`;
                chomp ($num2);
                my $num3 = 0;
                open (NUM, "$newdir/constrdir_res/num$i.txt") || die ("Cannot open $newdir/constrdir_res/num$i.txt");
                while (my $nline = <NUM>)
                {
                    chomp ($nline);
                    $num3 += $nline;
                }
                close(NUM);
                # now we must have this: $num1==$num2==$num3
                if ($num1 != $num2)
                {
                    print "ERROR! Num rows in $newdir/constrdir_res/constr$i.txt is $num1, which is different from num rows in $newdir/constrdir_res/constr_ub$i.txt that is $num2!";
                    print OPT "ERROR! Num rows in $newdir/constrdir_res/constr$i.txt is $num1, which is different from num rows in $newdir/constrdir_res/constr_ub$i.txt that is $num2!";
                    exit;
                }
                if ($num1 != $num3)
                {
                    print "ERROR! Num rows in $newdir/constrdir_res/constr$i.txt is $num1, which is different from the sum of values in $newdir/constrdir_res/num$i.txt that is $num3!";
                    print OPT "ERROR! Num rows in $newdir/constrdir_res/constr$i.txt is $num1, which is different from the sum of values in $newdir/constrdir_res/num$i.txt that is $num3!";
                    exit;
                }

                system ("cat $newdir/constrdir_res/constr$i.txt >> $struct_constr_file");
                system ("cat $newdir/constrdir_res/constr_ub$i.txt >> $struct_constr_ub_file");
                system ("cat $newdir/constrdir_res/num$i.txt >> $num_constraints_permol_filename");
            }

        }
        #$num_struct_constr = `egrep -E "sc[0-9]+:" $newdir/tmp | wc -l`;
        # The above is not good when I do the constraints in parallel, because there I don't have names for the constraints
        # When solver is IPOPT, I don't have any equality or inequality signs
        if ($solver eq "IPOPT")
        {
            $num_struct_constr = `cat $struct_constr_file | wc -l`;
        }
        else
        {
            $num_struct_constr = `cat $struct_constr_file | grep "[<>=]" | wc -l`;
        }
        chomp ($num_struct_constr);
        #print "Num str constr: $num_struct_constr\n";
        
        # now read the info from $num_constraints_permol_filename
        $num_molecules = 0;
        open (NUM, $num_constraints_permol_filename) || die ("Cannot open file $num_constraints_permol_filename");
        my $numline;
        while ($numline = <NUM>)
        {
            chomp($numline);
            $num_constraints_permol[$num_molecules] = $numline;
            $num_molecules++;
        }
        close (NUM);
    }

    ####################################################
    # in case $solver is IPOPT, I have to write the CPP file that implements the problem class
    # IPOPT is used if the predictor is HotKnots
    if ($solver eq "IPOPT")
    {

        my $regularizer_constraint = "";
        my $regularizer_upper_bound;
        my @means = ();
        # added this on Apr 9, 2008, and modified to add it as constraint on June 5, 2008
        if ($add_regularizer == 1)  {    die ("DIDN't implement lasso regularizer!\n");  }
        if ($add_regularizer == 2)
        {
            # first get the mean values            
            my $mline;
            my $coefficient;
            $there_are_non_zero_means = 0;
            open (MEANS, "$directory/$regularizer_mean_file") || die ("Cannot open $directory/$regularizer_mean_file");
            while ($mline = <MEANS>)
            {
                chomp ($mline);
                if ($mline != 0)    {   $there_are_non_zero_means = 1;  }
                push (@means, $mline);
            }
            close (MEANS);
            # now we also implement non-zero means
            #if ($there_are_non_zero_means)  {   die ("Non-zero regularizer means not implemented yet!\n");   }
            
            # sanity check
            if (@means."" != $num_params)
            {
                die ("The regularizer mean file does not have $num_params values!\n");
            }

            my $i;
            for ($i=0; $i < $num_params; $i++)
            {
                $regularizer_constraint .= "x[$i]*x[$i] ";
                if ($i % 6 == 5)
                { $regularizer_constraint .= "\n        "; }
                if ($i < $num_params-1)
                {   $regularizer_constraint .= " + ";  }
            }
            # add the non-zero means
            for ($i=0; $i < $num_params; $i++)
            {
                if ($means[$i] != 0)
                {
                    $regularizer_constraint .= " - 2*($means[$i])*x[$i]";
                    if ($i % 6 == 5)
                    { $regularizer_constraint .= "\n        "; }
                }
            }

            # the right hand side
            # added the normalization on June 27, 2008
            #$regularizer_upper_bound = $regularizer_bound_eta*($num_params-$num_zero_params);

            # forget about the zero params!!!!
            $regularizer_upper_bound = $regularizer_bound_eta*$num_params;
            # if non-zero mean, to add something here
            for ($i=0; $i < $num_params; $i++)
            {
                $regularizer_upper_bound -= $means[$i]*$means[$i];
            }

            # normalize this inequality, so that eta doesn't depend on the number of parameters in the model
            # the given eta should be small (around 1 or so)            
            # done writting the regularizer constraint
        }
    
        $ipopt_cpp = "$newdir/qcp.cpp";
        my $cname = "QCP";  # the class name
        # The order of the constraints is: structural, thermo, and additional
        my $num_constraints = $num_struct_constr + $num_thermo_constr + $num_additional_constr;
        if ($add_regularizer == 2)  { $num_constraints++;   }
        # There's a variable for each: energy parameter, structural constraint and thermo constraint, and this is the order
        my $num_variables = $num_params + $num_struct_constr + $num_thermo_constr;
        my $num_nonzero_jacobian = 0;   # number of nonzero elements in the jacobian
        my $num_nonzero_hessian = 0;    # number of nonzero elements in the hessian of the Lagrangian
        open (IPOPT, ">$ipopt_cpp") || die ("Cannot open file $ipopt_cpp");
        print IPOPT "#include \"qcp.hpp\"\n\n";
        print IPOPT "using namespace Ipopt;\n\n";
        print IPOPT "// constructor\n${cname}::${cname}()\n{}\n\n";
        print IPOPT "// destructor\n${cname}::~${cname}()\n{}\n\n";

        ##################
        # write the function get_starting_point
        # the starting point is in file $params_file
        open (INIT, "$params_file") || die ("Cannot open $params_file");
        print IPOPT "bool ${cname}::get_starting_point (Index n, bool init_x, Number* x,\n";
        print IPOPT "                              bool init_z, Number* z_L, Number* z_U,\n";
        print IPOPT "                              Index m, bool init_lambda, Number* lambda)\n";
        print IPOPT "{\n";
        print IPOPT "    assert (n == $num_variables);\n";
        print IPOPT "    assert (init_x == true);\n";
        print IPOPT "    assert (init_z == false);\n";
        print IPOPT "    assert (init_lambda == false);\n\n";
        print IPOPT "    // initialize to the given starting point\n";
        my $i=0;
        while ($line = <INIT>)
        {
            chomp ($line);
            print IPOPT "    x[$i] = $line;\n";
            $i++;
        }
        #sanity check
        $i == $num_params || die ("ERROR! File with initial params $params_file has $i entries, and not $num_params!\n");
        # the other variables are the variables corresponding to the structural and thermo constraints
        while ($i < $num_variables)
        {
            # for now just initialize with 1. TODO: check if other values are better
            # I had 1 before
            print IPOPT "    x[$i] = 0.0;\n";
            $i++;
        }
        print IPOPT "    return true;\n";        
        print IPOPT "}\n\n";
        close (INIT);

        ##################
        # write the function get_bounds_info
        print IPOPT "bool ${cname}::get_bounds_info (Index n, Number* x_l, Number* x_u,\n";
        print IPOPT "                           Index m, Number* g_l, Number* g_u)\n";
        print IPOPT "{\n";
        print IPOPT "    assert (n == $num_variables);\n";
        print IPOPT "    assert (m == $num_constraints);\n\n";

        # First, the bounds on the energy parameters
        print IPOPT "    // bounds on x\n";
        # these are in @params_lb and @params_ub;
        for ($i=0; $i <= $#params_lb; $i++)
        {
            if ($fixparams_file eq "none")
            {
                print IPOPT "    x_l[$i] = $params_lb[$i];    x_u[$i] = $params_ub[$i];\n";
            }
            else
            {
                if ($fixparams[$i] eq "variable")
                {
                    print IPOPT "    x_l[$i] = $params_lb[$i];    x_u[$i] = $params_ub[$i];\n";
                }
                else
                {
                    print IPOPT "    x_l[$i] = $fixparams[$i];    x_u[$i] = $fixparams[$i];\n";
                }
            }
        }
        # just added the variables for the energy parameters
        # sanity check
        $i == $num_params || die ("ERROR! Number of bounds on energy parameters is $i and not $num_params!\n");
        # now add the variables for the structural constraints. These have to be positive if !$max_margin
        for (my $i=$num_params; $i < $num_params + $num_struct_constr; $i++)
        {
            if ($max_margin == 1 || $max_margin == 2)   
            {
                print IPOPT "    x_l[$i] = -2e19;    x_u[$i] = 2e19;\n";
            }
            else    
            {
                print IPOPT "    x_l[$i] = 0.0;    x_u[$i] = 2e19;\n";
            }
        }
        # now add the variables for the thermo constraints. These are unbounded
        for (my $i=$num_params+$num_struct_constr; $i < $num_params + $num_struct_constr + $num_thermo_constr; $i++)
        {
            print IPOPT "    x_l[$i] = -2e19;    x_u[$i] = 2e19;\n";
        }

        # Now add the bounds of the constraints
        # First, the bounds of the structural constraints
        # The lower bounds are -inf for all structural constraints, by definition
        # The upper bounds are given in the file $struct_constr_ub_file
        $index=0;
        open (UB, $struct_constr_ub_file) || die ("Cannot open file $struct_constr_ub_file");
        while ($line = <UB>)
        {
            chomp ($line);
            if ($max_margin == 1)    # in this case we have equality constraints
            {
                print IPOPT "    g_l[$index] = $line;    g_u[$index] = $line;\n";
            }
            else    # if ($max_margin == 0 || $max_margin == 2)     # inequality constraints
            {
                print IPOPT "    g_l[$index] = -2e19;    g_u[$index] = $line;\n";
            }
            $index++;
        }
        close (UB);
        # sanity check
        $index == $num_struct_constr || die ("ERROR! Num lines in $struct_constr_ub_file is not $num_struct_constr");

        # Next, the bounds of the thermo constraints. These are in the array @thermo_bounds
        # first do sanity check
        @thermo_bounds."" == $num_thermo_constr || die ("ERROR! Num values of thermo_bounds is " .@thermo_bounds.", not $num_thermo_constr");
        for (my $i=0; $i < $num_thermo_constr; $i++)
        {
            print IPOPT "    g_l[".($num_struct_constr+$i)."] = $thermo_bounds[$i];    g_u[".($num_struct_constr+$i)."] = $thermo_bounds[$i];\n";
        }
        # Next, the bounds of the additional constraints
        # CAUTION: I'm assuming these are 0, but I'm not checking any file. TO ADD such a file if needed!
        # Now I added a lower bound and an upper bound file. If they are given, use that, otherwise use 0.
        if ($additional_constraints_lb_file ne "none")
        {
            open (LB, "$directory/$additional_constraints_lb_file") || die ("Cannot open $directory/$additional_constraints_lb_file");
            my $i = 0;
            while ($line = <LB>)
            {
                chomp ($line);
                print IPOPT "    g_l[".($num_struct_constr+$num_thermo_constr+$i)."] = $line;\n";
                $i++;
            }                
            close (LB);
            $i == $num_additional_constr || die ("Number of rows in $directory/$additional_constraints_lb_file is $i and should be $num_additional_constr! ABORT!");
        }
        else
        {
            for (my $i=0; $i < $num_additional_constr; $i++)
            {
                print IPOPT "    g_l[".($num_struct_constr+$num_thermo_constr+$i)."] = 0;\n";
            }
        }

        if ($additional_constraints_ub_file ne "none")
        {
            open (UB, "$directory/$additional_constraints_ub_file") || die ("Cannot open $directory/$additional_constraints_ub_file");
            my $i = 0;
            while ($line = <UB>)
            {
                chomp ($line);
                print IPOPT "    g_u[".($num_struct_constr+$num_thermo_constr+$i)."] = $line;\n";
                $i++;
            }                
            close (UB);
            $i == $num_additional_constr || die ("Number of rows in $directory/$additional_constraints_ub_file is $i and should be $num_additional_constr! ABORT!");
        }
        else
        {
            for (my $i=0; $i < $num_additional_constr; $i++)
            {
                print IPOPT "    g_u[".($num_struct_constr+$num_thermo_constr+$i)."] = 0;\n";
            }
        }
        # add bounds for the regularizer constraint
        if ($add_regularizer == 2)
        {
            print IPOPT "    g_l[".($num_struct_constr+$num_thermo_constr+$num_additional_constr)."] = -2e19;";
            print IPOPT "    g_u[".($num_struct_constr+$num_thermo_constr+$num_additional_constr)."] = $regularizer_upper_bound;\n";
        }
        print IPOPT "    return true;\n";
        print IPOPT "}\n\n";

        ##################
        # write the function eval_f
        print IPOPT "bool ${cname}::eval_f (Index n, const Number* x, bool new_x, Number& obj_value)\n";
        print IPOPT "{\n";
        print IPOPT "    assert (n == $num_variables);\n";        
        print IPOPT "    obj_value = ";
        if ($normalize ne "n2" || $a_param != 2 || $lendep eq "lendep-yes")
        {   print "NOT IMPLEMENTED! ABORT! normalize must be n2 and a_param must be 2!\n";  exit;   }

        # also write the gradient coefficients, to have them for the function eval_grad_f
        my @gradient_coefficient = ();
        for (my $i=0; $i < $num_variables; $i++)    {   $gradient_coefficient[$i] = 0;  }
        $unique_index = 0;
        if ($train_file ne "none")
        {
            for (my $perm=0; $perm < $num_molecules; $perm++)
            {
                # TODO: figure out the deltas for the case $max_margin == 2
                # for $max_margin == 2, we have one one delta per molecule that has constraints
                my $likelihood_normalizer;
                if ($num_constraints_permol[$perm] != 0)
                {
                    if ($normalize_all_three_sums)
                    {
                        $likelihood_normalizer = $scaler/($num_constraints_permol[$perm]*$num_molecules);
                    }
                    else
                    {
                        $likelihood_normalizer = $scaler/$num_constraints_permol[$perm];
                    }
                }
                for (my $j=0; $j < $num_constraints_permol[$perm]; $j++)
                {
                    # For $max_margin 0 and 1, we have one delta per constraint
                    if ($unique_index > 0)   {   print IPOPT " + "; }
                    if ($max_margin == 1)
                    {
                        if ($dimcg_exponent == 1)
                        {
                            printf IPOPT (" %e*x[%d] ", $likelihood_normalizer, $num_params+$unique_index);
                            $gradient_coefficient[$num_params+$unique_index] = $likelihood_normalizer;
                        }
                        else        # it's 2  # This doesn't work, the problem is not differentiable
                        {
                            my $ind = $num_params+$unique_index;
                            printf IPOPT (" %e*x[%d]*fabs(x[%d]) ", $likelihood_normalizer, $ind, $ind);
                            $gradient_coefficient[$ind] = (2*$likelihood_normalizer) . " * fabs(x[$ind])";
                        }                        
                    }
                    else # if ($max_margin == 0)
                    {
                        printf IPOPT (" %e*x[%d]*x[%d] ", $likelihood_normalizer, $num_params+$unique_index, $num_params+$unique_index);
                        $gradient_coefficient[$num_params+$unique_index] = $likelihood_normalizer*2;
                    }                    
                    if ($unique_index % 4 == 3)
                    { print IPOPT "\n        "; }
                    $unique_index++;
                }
            }
            #sanity check
            $unique_index == $num_struct_constr || die ("ERROR! unique_index=$unique_index and should be $num_struct_constr!");            
        }
        
        if ($lambda > 0)    # try to emphasise more the thermodynamic measurements, by multiplying with $num_thermo_constr
        {
            # add the v variable, corresponding to the thermodynamic constraints
            # the v variable is the xi variable in the ISMB paper
            for (my $i=0; $i < $num_thermo_constr; $i++)
            {
                # Until Apr 17, 2008, I was normalizing by the number of thermo experiments, as follows:
                # my $thermo_normalizer =  $lambda * 1000/$num_thermo_constr * $thermo_precisions[$i];

                # Starting Apr 17, 2008, I'm not dividing by the number of experiments any more
                my $thermo_normalizer;
                if ($normalize_all_three_sums)
                {
                    $thermo_normalizer =  $lambda * $scaler * $thermo_precisions[$i] / $num_thermo_constr;
                }
                else
                {
                    $thermo_normalizer =  $lambda * $scaler * $thermo_precisions[$i];
                }
                if ($i==0)   {   printf IPOPT ("\n         + ")   ;   }
                else         {   printf IPOPT (" + ")   ;   }
                printf IPOPT ("%e*x[%d]*x[%d] ", $thermo_normalizer, $num_params+$num_struct_constr+$i, $num_params+$num_struct_constr+$i);
                $gradient_coefficient[$num_params+$num_struct_constr+$i] = $thermo_normalizer*2;
                if ($i % 4 == 3)
                { print IPOPT "\n        "; }
            }
        }
        
        print IPOPT ";\n    return true;\n";
        print IPOPT "}\n\n";


        ##################
        # write the function eval_grad_f
        print IPOPT "bool ${cname}::eval_grad_f (Index n, const Number* x, bool new_x, Number* grad_f)\n";
        print IPOPT "{\n";
        print IPOPT "    assert (n == $num_variables);\n";
        for (my $i=0; $i < $num_variables; $i++)
        {
            # Not sure if I shouldn't write the 0 ones. I'm writting them, just in case.
            #if ($gradient_coefficient[$i] != 0)
            # I'm adding the case max_margin
            if ($i >= $num_params && $i < $num_params + $num_struct_constr && ($max_margin == 1 || $max_margin == 2))
            # TODO: for max_margin == 2, we want sum of delta, not sum of delta^2
            {
                print IPOPT "    grad_f[$i] = $gradient_coefficient[$i];\n";
            }
            else #  if ($max_margin == 0)
            {
                print IPOPT "    grad_f[$i] = $gradient_coefficient[$i] * x[$i];\n";
            }
        }
        print IPOPT "    return true;\n";
        print IPOPT "}\n\n";

        ##################
        # write the function eval_g
        print IPOPT "bool ${cname}::eval_g (Index n, const Number* x, bool new_x, Index m, Number* g)\n";
        print IPOPT "{\n";
        print IPOPT "    assert (n == $num_variables);\n";
        print IPOPT "    assert (m == $num_constraints);\n";
                                
        $index = 0;
        # First, add the structural constraints from the file $struct_constr_ub_file                                
        open (STR, $struct_constr_file) || die ("Cannot open file $struct_constr_file");
        while ($line = <STR>)
        {
            chomp ($line);
            print IPOPT "    g[$index] = $line - x[" . ($num_params+$index) . "];\n";
            $index++;
        }
        close (STR);
        # sanity check
        $index == $num_struct_constr || die ("ERROR! Number of constraints in $struct_constr_file is $index and should be $num_struct_constr!\n");

        # Next, add the thermodynamic constraints
        # continue with $index
        if ($lambda > 0)
        {
            open (THERMO, $thermo_constraints_file) || die ("Cannot open file $thermo_constraints_file");
            while ($line = <THERMO>)
            {
                chomp ($line);
                print IPOPT "    g[$index] = $line - x[". ($num_params+$index) ."];\n";
                $index++;
            }
            close (THERMO);
            # sanity check
            $index == $num_struct_constr + $num_thermo_constr || die ("ERROR! Number of constraints in struct and thermo should be " .($num_struct_constr + $num_thermo_constr).", and is $index!");
        }

        # Next, add the additional constraints
        # continue with $index
        if ($additional_constraints_file ne "none")
        {
            open (ADD, "$directory/$additional_constraints_file") || die ("Cannot open $directory/$additional_constraints_file");
            while ($line = <ADD>)
            {
                chomp ($line);
                print IPOPT "    g[$index] = $line;\n";      
                $index++;
            }
            close (ADD);
            # sanity check
            $index == $num_struct_constr + $num_thermo_constr + $num_additional_constr || die ("ERROR! Number of constraints in struct, thermo and additional should be " .($num_struct_constr + $num_thermo_constr+ $num_additional_constr).", and is $index!");
        }
        
        if ($add_regularizer == 2)
        {
            print IPOPT "    g[$index] = $regularizer_constraint;\n";
        }
        print IPOPT "    return true;\n";
        print IPOPT "}\n\n";
        
        ##################
        # write the function eval_jac_g
        # Get the jacobian by using $struct_constr_file
        print IPOPT "bool ${cname}::eval_jac_g (Index n, const Number* x, bool new_x,\n";
        print IPOPT "                      Index m, Index nele_jac, Index* iRow, Index *jCol,\n";
        print IPOPT "                      Number* values)\n";
        print IPOPT "{\n";
        print IPOPT "    assert (n == $num_variables);\n";
        print IPOPT "    assert (m == $num_constraints);\n";
        #print IPOPT "    assert (nele_jac == );\n";

        # fill up the values of the jacobian of the constraints first
        # STRATEGY is: look at all the constraints and do the symbolic differentiation.
        my @irow = ();
        my @jcol = ();
        my $jacobian = "";
        print IPOPT "    if (values != NULL)\n";
        print IPOPT "    {\n";        
        # start with structural constraints
        # The Jacobian has $num_constraints rows and $num_variables columns
        my $row = 0;     # the constraint number, i.e. the row in the Jacobian
        open (STR, $struct_constr_file) || die ("Cannot open file $struct_constr_file");
        while ($line = <STR>)
        {
            chomp ($line);
            #for (my $col = 0; $col < $num_variables; $col++)
            # I don't need to go all the way to $num_variables, it's enough to go to $num_params
            for (my $col = 0; $col < $num_params; $col++)
            {
                $jacobian = get_derivative ($line, $col);
                if ($jacobian ne "")    # value is non-0, store it
                {                    
                    push (@irow, $row);
                    push (@jcol, $col);
                    print IPOPT "        values[$num_nonzero_jacobian] = $jacobian;  // $row,$col\n";
                    $num_nonzero_jacobian++;
                }
            }
            # NOW don't forget to add the derivative of the "d" (slack) variable, which is always -1.
            push (@irow, $row);
            push (@jcol, $num_params+$row);
            print IPOPT "        values[$num_nonzero_jacobian] = -1;  // $row,". ($num_params+$row) ."\n";
            $num_nonzero_jacobian++;
            $row++;
        }
        close (STR);

        # Add thermo constraints
        # continue with $row and with $num_nonzero_jacobian
        if ($lambda > 0)
        {
            open (THERMO, $thermo_constraints_file) || die ("Cannot open file $thermo_constraints_file");
            while ($line = <THERMO>)
            {
                chomp ($line);            
                # for (my $col = 0; $col < $num_variables; $col++)
                # I don't need to go all the way to $num_variables, it's enough to go to $num_params
                for (my $col = 0; $col < $num_params; $col++)
                {
                    $jacobian = get_derivative ($line, $col);
                    if ($jacobian ne "")    # value is non-0, store it
                    {                    
                        push (@irow, $row);
                        push (@jcol, $col);
                        print IPOPT "        values[$num_nonzero_jacobian] = $jacobian;  // $row,$col\n";
                        $num_nonzero_jacobian++;
                    }
                }
                # NOW don't forget to add the derivative of the "v" (slack) variable, which is always -1.
                push (@irow, $row);
                push (@jcol, $num_params+$row);
                print IPOPT "        values[$num_nonzero_jacobian] = -1;  // $row,". ($num_params+$row) ."\n";
                $num_nonzero_jacobian++;
                $row++;
            }
            close (THERMO);
        }
        
        # Add additional constraints
        # continue with $row and with $num_nonzero_jacobian
        if ($additional_constraints_file ne "none")
        {
            open (ADD, "$directory/$additional_constraints_file") || die ("Cannot open file $directory/$additional_constraints_file");
            while ($line = <ADD>)
            {
                chomp ($line);
                for (my $col = 0; $col < $num_variables; $col++)
                {
                    $jacobian = get_derivative ($line, $col);
                    if ($jacobian ne "")    # value is non-0, store it
                    {                    
                        push (@irow, $row);
                        push (@jcol, $col);
                        print IPOPT "        values[$num_nonzero_jacobian] = $jacobian;  // $row,$col\n";
                        $num_nonzero_jacobian++;
                    }
                }
                # I have no slack variable here
                $row++;
            }
            close (ADD);
        }
        
        # Add the regularizer constraint
        if ($add_regularizer == 2)
        {
            for (my $col = 0; $col < $num_params; $col++)   # all params are squared
            {
                push (@irow, $row);
                push (@jcol, $col);
                if ($means[$col] == 0)
                {
                    print IPOPT "        values[$num_nonzero_jacobian] = 2 * x[$col];  // $row,$col\n";
                }
                else
                {
                    print IPOPT "        values[$num_nonzero_jacobian] = 2 * x[$col] - 2 * ($means[$col]);  // $row,$col\n";
                }
                $num_nonzero_jacobian++;
            }
            $row++;
        }
                
        print IPOPT "    }\n";
        print IPOPT "    else\n";
        print IPOPT "    {\n";        
        # fill up the structure of the Jacobian
        for (my $i=0; $i < $num_nonzero_jacobian; $i++)
        {               
            print IPOPT "        iRow[$i] = $irow[$i];    jCol[$i] = $jcol[$i];\n";
        }
        print IPOPT "    }\n";
        print IPOPT "    return true;\n";
        print IPOPT "}\n\n";

        ##################
        # write the function eval_h
        # Get the Hessian of the Lagrangian L, where L = obj_factor * f(x) + g(x)' lambda
        print IPOPT "bool ${cname}::eval_h (Index n, const Number* x, bool new_x,\n";
        print IPOPT "                  Number obj_factor, Index m, const Number* lambda,\n";
        print IPOPT "                  bool new_lambda, Index nele_hess, Index* iRow,\n";
        print IPOPT "                  Index* jCol, Number* values)\n";
        print IPOPT "{\n";
        print IPOPT "    assert (n == $num_variables);\n";
        print IPOPT "    assert (m == $num_constraints);\n";
        #  // ignore new_x
        #  // ignore new_lambda
        # assert (nele_hess == 10);

        # DON'T use a matrix to store every hessian value, because it ca run out of space!
        
        my @hessian_index = ();       # hessian of the lagrangian.
        # Now none of the elements of hessian_index "exists"
        my $current_index = 0;

        print IPOPT "    if (values != NULL)\n";
        print IPOPT "    {\n";
                
        # Only the non-zero values in the lower left triangle are stored in this function
        for (my $row=0; $row < $num_variables; $row++)  # the hessian row
        {
            # First, write the objective part. Here, we have a weighted sum of squares, so there's something there only when $row=$col
            my $col = $row;
            # In case of max_margin, this term in the objective is linear, so there's nothing to add here
            if ($row >= $num_params && $row < $num_params + $num_struct_constr && ($max_margin == 1 || $max_margin == 2))
            # TODO: figure out $max_margin == 2
            {
                if ($dimcg_exponent == 1)
                {
                    next;   # nothing to add
                }
                else
                {
                    if (! exists($hessian_index[$row][$col]))
                    {
                        $hessian_index[$row][$col] = $current_index++;
                        print IPOPT "        values[$hessian_index[$row][$col]] = obj_factor * $gradient_coefficient[$row]/x[$row];   // $row,$col\n";
                    }
                    else
                    {
                        print IPOPT "        values[$hessian_index[$row][$col]] += obj_factor * $gradient_coefficient[$row]/x[$row];   // $row,$col\n";
                    }                    
                    next;
                }
            }
            if ($gradient_coefficient[$row] != 0)
            {
                if (! exists($hessian_index[$row][$col]))
                {
                    $hessian_index[$row][$col] = $current_index++;
                    print IPOPT "        values[$hessian_index[$row][$col]] = obj_factor * $gradient_coefficient[$row];   // $row,$col\n";
                }
                else
                {
                    print IPOPT "        values[$hessian_index[$row][$col]] += obj_factor * $gradient_coefficient[$row];   // $row,$col\n";
                }
            }
        }

        # Second, write the structural constraints part
        open (STR, $struct_constr_file) || die ("Cannot open file $struct_constr_file");
        my $der2;   # second derivative
        $index = 0;     # the constraint index
        while ($line = <STR>)
        {
            chomp ($line);
            #for (my $row=0; $row < $num_variables; $row++)  # the hessian row
            for (my $row=0; $row < $num_params; $row++)
            {
                $jacobian = get_derivative ($line, $row);
                if ($jacobian ne "")    # value is non-0, continue to get the second derivative
                {
                    for (my $col=0; $col <= $row; $col++)
                    {
                        $der2 = get_derivative ($jacobian, $col);
                        # $der2 is one number, eliminate all the spaces
                        $der2 =~ s/\s//g;
                        if ($der2 ne "")    # value is non-0, we have something for the hessian of the Lagrangian
                        {
                            if (! exists($hessian_index[$row][$col]))
                            {
                                $hessian_index[$row][$col] = $current_index++;
                                print IPOPT "        values[$hessian_index[$row][$col]] = lambda[$index] * $der2;   // $row,$col\n";
                            }
                            else
                            {
                                print IPOPT "        values[$hessian_index[$row][$col]] += lambda[$index] * $der2;   // $row,$col\n";
                            }
                        }
                    }
                }
            }
            $index++;
        }
        close (STR);                
                
        # CAUTION! Write the thermo constraints part IN CASE IT BECOMES QUADRATIC, otherwise I'm done.
        # Doing this now, after I added a few experiments on pseudoknots
        if ($lambda > 0)
        {
            open (THERMO, $thermo_constraints_file) || die ("Cannot open file $thermo_constraints_file");
            # continue with the constraint index $index
            while ($line = <THERMO>)
            {
                chomp ($line);            
                # I don't need to go all the way to $num_variables, it's enough to go to $num_params
                for (my $row=0; $row < $num_params; $row++)
                {
                    $jacobian = get_derivative ($line, $row);
                    if ($jacobian ne "")    # value is non-0, continue to get the second derivative
                    {            
                        for (my $col = 0; $col <= $row; $col++)
                        {
                            $der2 = get_derivative ($jacobian, $col);
                            # $der2 is one number, eliminate all the spaces
                            $der2 =~ s/\s//g;
                            if ($der2 ne "")    # value is non-0, we have something for the hessian of the Lagrangian
                            {
                                if (! exists($hessian_index[$row][$col]))
                                {
                                    $hessian_index[$row][$col] = $current_index++;
                                    print IPOPT "        values[$hessian_index[$row][$col]] = lambda[$index] * $der2;   // $row,$col\n";
                                }
                                else
                                {
                                    print IPOPT "        values[$hessian_index[$row][$col]] += lambda[$index] * $der2;   // $row,$col\n";
                                }
                            }
                        }
                    }
                }
                $index++;
            }
            close (THERMO);
        }
        
        # CAUTION! Write the additional constraints part  IN CASE IT BECOMES QUADRATIC, otherwise I'm done.
        $index += $num_additional_constr;

        if ($add_regularizer == 2)
        {
            for (my $row=0; $row < $num_params; $row++)
            {
                my $col = $row;
                if (! exists($hessian_index[$row][$col]))
                {
                    $hessian_index[$row][$col] = $current_index++;
                    print IPOPT "        values[$hessian_index[$row][$col]] = lambda[$index] * 2;   // $row,$col\n";
                }
                else
                {
                    print IPOPT "        values[$hessian_index[$row][$col]] += lambda[$index] * 2;   // $row,$col\n";
                }
            }
            $index++;
        }

        print IPOPT "    }\n";
        $num_nonzero_hessian = $current_index;
        # now write down the hessian nonzero structure
        print IPOPT "    else\n";
        print IPOPT "    {\n";
        for (my $row=0; $row < $num_variables; $row++)  # the hessian row
        {
            for (my $col=0; $col <= $row; $col++)
            {
                if ( exists ($hessian_index[$row][$col]))     # yay, there's a value here
                {                    
                    print IPOPT "        iRow[$hessian_index[$row][$col]] = $row;     jCol[$hessian_index[$row][$col]] = $col;\n";
                }
            }
        }
        print IPOPT "    }\n";
        print IPOPT "    return true;\n";
        print IPOPT "}\n\n";

  
        ##################
        # write the function get_nlp_info
        print IPOPT "bool ${cname}::get_nlp_info (Index& n, Index& m, Index& nnz_jac_g,\n";
        print IPOPT "                        Index& nnz_h_lag, IndexStyleEnum& index_style)\n";
        print IPOPT "{\n";
        print IPOPT "    n = $num_variables;\n";
        print IPOPT "    m = $num_constraints;\n";
        print IPOPT "    nnz_jac_g = $num_nonzero_jacobian;\n";
        print IPOPT "    nnz_h_lag = $num_nonzero_hessian;\n";
        print IPOPT "    index_style = TNLP::C_STYLE;\n";
        print IPOPT "    return true;\n";
        print IPOPT "}\n\n";

        ##################
        # write the function finalize_solution
        print IPOPT "void ${cname}::finalize_solution(SolverReturn status,\n";
        print IPOPT "                            Index n, const Number* x, const Number* z_L, const Number* z_U,\n";
        print IPOPT "                            Index m, const Number* g, const Number* lambda,\n";
        print IPOPT "                            Number obj_value,\n";
        print IPOPT "                            const IpoptData* ip_data,\n";
        print IPOPT "                            IpoptCalculatedQuantities* ip_cq)\n";
        print IPOPT "{\n";
        print IPOPT "  // For this example, we write the solution to the console\n";
        print IPOPT "  printf(\"\\n\\nSolution of the primal variables, x\\n\");\n";
        print IPOPT "  for (Index i=0; i<n; i++) {\n";
        print IPOPT "    printf(\"x[%d] = %e\\n\", i, x[i]);\n";
        print IPOPT "  }\n";
        print IPOPT "  printf(\"\\n\\nSolution of the bound multipliers, z_L and z_U\\n\");\n";
        print IPOPT "  for (Index i=0; i<n; i++) {\n";
        print IPOPT "    printf(\"z_L[%d] = %e\\n\", i, z_L[i]);\n";
        print IPOPT "  }\n";
        print IPOPT "  for (Index i=0; i<n; i++) {\n";
        print IPOPT "    printf(\"z_U[%d] = %e\\n\", i, z_U[i]);\n";
        print IPOPT "  }\n";
        print IPOPT "  printf(\"\\n\\nObjective value\\n\");\n";
        print IPOPT "  printf(\"f(x*) = %e\\n\", obj_value);\n";
        print IPOPT "  printf(\"\\nFinal value of the constraints:\\n\");\n";
        print IPOPT "  for (Index i=0; i<m ;i++) {\n";
        print IPOPT "    printf(\"g(%d) = %e\\n\", i, g[i]);\n";
        print IPOPT "  }\n";
        print IPOPT "}\n\n\n";
        close (IPOPT);

        # now I have all the files for IPOPT
        #system ("cd $newdir; make -f Makefile_qcp");
        # make takes a lot of memory, try to send it to sikinos, which has 4GB
        # I don't know why skopelos or euler don't work
        # sikinos doesn't respond
        #system ("nohup ssh sikinos \"cd $newdir; make -f Makefile_qcp\"");
        #system ("nohup ssh serifos \"cd $newdir; make -f Makefile_qcp\"");
        
        # NOTE: If you don't want to run it on saria, but on the local machine, replace this line with:
        # system ("make -f Makefile_qcp");
        system ("nohup ssh saria \"cd $newdir; make -f Makefile_qcp\"");

        $answer_file = "$newdir/ipopt_answer_${s}.txt";
        system ("$newdir/qcp_main > $answer_file");
        #print OPT "Starting qcp_main job on saria.cs.ubc.ca\n";
        #system ("nohup ssh saria \"$newdir/qcp_main > $answer_file");
        parse_IPOPT_answer ($answer_file, $new_params_file);
    }   # finished with solver IPOPT
    
    ####################################################

    elsif ($solver eq "CPLEX")
    {
        $problem_file_nopath = "problem_sub-${s}.lp";
        $problem_file = "$newdir/$problem_file_nopath";
        
        $prev_basis_file = "$newdir/basis_sub-${tmp}.bas";
        if ($opt_problem eq "lp")
        {
            $answer_file_nopath = "answer_sub-${s}.txt";
        }
        else    # qp
        {
            $answer_file_nopath = "answer_sub-${s}.sol";
        }     
        $answer_file = "$newdir/$answer_file_nopath";

        ##################################################
        # write the objective function
        ##################################################
        
        open (PROB, ">$problem_file") || die ("Cannot open $problem_file");
        print PROB "MINIMIZE\n";        
        
        # add the d variables, corresponding to the structural constraints
        
 
        if ($lendep eq "lendep-yes")
        {           
            # add dependence on the sequence length
            open (DATA, $datafile) || die ("Cannot open $datafile");        
            $truestr = "";
            $index = 0;
        }    
        #print "Num sc: $num_struct_constr\n";
        
        
        # instead of normalizing everything by the total number pf constraints, normalize the constraints of one molecule by the number of constraints for that molecule, and then normalize by the number of molecules.        

        if ($normalize eq "n1")     # by default, $normalize is n2, seems to be better
        {
            # outdated
            for (my $i=0; $i < $num_struct_constr; $i++)
            {
                if ($lendep eq "lendep-yes")
                {
                    # get the length of the sequence
                    $strlen = 0;
                    while ($line = <DATA>)
                    {
                        if ($line =~ /^[\n>]/) { next; }
                        chomp ($line);
                        #print "*** $line\n";
                        if ($line =~ /^[ACGUTacgut]+$/)
                        {
                            $truestr = <DATA>;
                            chomp ($truestr);
                            next;
                        }
                        # the remaining lines are predicted structures
                        if ($line eq $truestr)   { next; }  # correct prediction
                        $strlen = length ($line);
                        $index ++;
                        last;
                    }
                    if ($strlen == 0) { print "ERRRORRR, strlen=0, index=$index, i=$i, line=|$line|\n"; exit; }
                }
                # now write the optimization function
                if ($a_param == 1)      # highly outdated, $a_param 2 is up-to-date
                {
                    if ($i == 0) { print PROB "    "; }
                    else         { print PROB " + ";  }
                    if ($lendep eq "lendep-yes")    # this is highly outdated
                    {
                        printf PROB (" %e d%d", (1-$lambda)*1000/$strlen, $i);
                    }
                    else
                    {
                        if ($lambda > 0)    #
                        {
                            printf PROB (" %e d%d", (1-$lambda)*1000/$num_struct_constr, $i);
                        }
                        else
                        {
                            printf PROB (" d%d", $i);
                        }
                    }
                    if ($i!=0 && $i % 10 == 0)
                    { print PROB "\n        "; }
                }
                else    # a_param is 2, up-to-date
                {
                    if ($i == 0) { print PROB "    [ "; }
                    else         { print PROB " + ";  }
                    if ($lendep eq "lendep-yes")    # highly outdated
                    {
                        printf PROB (" %e d%d*d%d", (1-$lambda)*1000/$strlen, $i, $i);
                    }
                    else
                    {
                        if ($lambda > 0)    #
                        {
                            #printf PROB (" %e d%d*d%d", (1-$lambda)*1000/$num_struct_constr, $i, $i);
                            printf PROB (" %e d%d^2", $scaler/$num_struct_constr, $i);
                        }
                        else
                        {
                            printf PROB (" d%d^2", $i);
                        }
                    }
                    if ($i != 0 && $i % 10 == 0)
                    { print PROB "\n        "; }
                }
            }
            
        }
        elsif ($normalize eq "n2")  # this is the default option, it should be up-to-date
        {
            if ($train_file ne "none" && $max_margin != 2 && ($max_margin != 1 || $max_margin == 1 && $dimcg_exponent == 2))
            {
                # now read the info from $num_constraints_permol_filename
                # The following is already done above
#                 $num_molecules = 0;
#                 open (NUM, $num_constraints_permol_filename) || die ("Cannot open file $num_constraints_permol_filename");
#                 my $numline;
#                 while ($numline = <NUM>)
#                 {
#                     chomp($numline);
#                     $num_constraints_permol[$num_molecules] = $numline;
#                     $num_molecules++;
#                 }
#                 close (NUM);

                $unique_index = 0;
                for (my $perm=0; $perm < $num_molecules; $perm++)
                {
                    my $likelihood_normalizer;
                    # Until Apr 17, 2008, the formula was (1-lambda)*normalized_likelihood + lambda*normalized_prior, as follows:
#                     if ($lambda > 0)
#                     {   $likelihood_normalizer = (1-$lambda)*1000/($num_constraints_permol[$perm]*$num_molecules);  }
#                     else
#                     {   $likelihood_normalizer = 1000/$num_constraints_permol[$perm];   }

                    # Starting Apr 17, 2008, we only use lambda to weigh the thermodynamic set, because now we also include regularizer
                    if ($num_constraints_permol[$perm] != 0)
                    {
                        if ($normalize_all_three_sums)
                        {
                            $likelihood_normalizer = $scaler/($num_constraints_permol[$perm]*$num_molecules);
                        }
                        else
                        {
                            $likelihood_normalizer = $scaler/$num_constraints_permol[$perm];
                        }
                    }
                    for (my $j=0; $j < $num_constraints_permol[$perm]; $j++)
                    {

                        # now write the optimization function
                        if ($a_param == 1)  # OUTDATED
                        {
                            if ($unique_index == 0) { print PROB "    "; }
                            else                    { print PROB " + ";  }
                            printf PROB (" %e d%d", $likelihood_normalizer, $unique_index);
                            if ($unique_index!=0 && $unique_index % 10 == 0)
                            { print PROB "\n        "; }
                        }
                        else
                        {
                            if ($unique_index == 0) { print PROB "    [ "; }
                            else                    { print PROB " + ";  }
                            if ($max_margin == 1 && $dimcg_exponent == 2)
                            {
                                # Should be d*abs(d), so I introduce another variable
                                printf PROB (" %e d%d*f%d", $likelihood_normalizer, $unique_index, $unique_index);
                            }
                            else
                            {
                                printf PROB (" %e d%d^2", $likelihood_normalizer, $unique_index);
                            }
                            if ($unique_index != 0 && $unique_index % 10 == 0)
                            { print PROB "\n        "; }
                        }
                        $unique_index++;
                    }
                }
            }
        }
        
        if ($lendep eq "lendep-yes")
        {   close (DATA);    }
        if ($train_file ne "none" && $a_param == 2 && $lambda == 0)  #  && $add_regularizer == 0 )
        # removed the regularizer, I'm adding it as constraint, not in the objective
        {
            print PROB " ] / 2";         
        }    
        
        if ($lambda > 0)    # try to emphasise more the thermodynamic measurements, by multiplying with $num_struct_constr
        {
            print PROB "\n        ";
            # add the v variable, corresponding to the thermodynamic constraints
            # the v variable is the xi variable in the ISMB paper
            if ($a_param == 2)
            {                
                for (my $i=0; $i < $num_thermo_constr; $i++)
                {
                    # Until Apr 17, 2008, I was normalizing by the number of thermo experiments, as follows:
                    # my $thermo_normalizer =  $lambda * 1000/$num_thermo_constr * $thermo_precisions[$i];

                    # Starting Apr 17, 2008, I'm not dividing by the number of experiments any more
                    my $thermo_normalizer;
                    if ($normalize_all_three_sums)
                    {
                        $thermo_normalizer =  $lambda * $scaler * $thermo_precisions[$i] / $num_thermo_constr;
                    }
                    else
                    {
                        $thermo_normalizer =  $lambda * $scaler * $thermo_precisions[$i];
                    }
                    if ($i==0 && ($train_file eq "none" || $max_margin == 2 || ($max_margin == 1 && $dimcg_exponent == 1)))
                    {   printf PROB ("       [ ")   ;   }
                    elsif ($i==0)                         {   printf PROB ("       + ")   ;   }
                    else                                    {   printf PROB (" + ")   ;   }
                    printf PROB ("%e v%d^2 ", $thermo_normalizer, $i);
                    #print PROB " + $num_struct_constr v$i*v$i";
                    if ($i % 10 == 0 && $i != 0)
                    { print PROB "\n        "; }            
                }
                #if (!$add_regularizer)     # adding the regularizer as constraint
                {
                    print PROB " ] / 2";
                }
            }
            else    # highly outdated
            {
                if ($train_file eq "none")
                {
                    printf PROB ("       [ %e w0", $lambda * 1000/$num_thermo_constr);
                }
                else
                {
                    printf PROB ("       + %e w0", $lambda * 1000/$num_thermo_constr);
                }
                #print PROB "       + $b_param * 1000/$num_thermo_constr w0 ";
                for (my $i=1; $i < $num_thermo_constr; $i++)
                {
                    printf PROB (" + %e w%d", $lambda * 1000/$num_thermo_constr, $i);
                    #print PROB " + $b_param * 1000/$num_thermo_constr w$i";
                    if ($i % 10 == 0)
                    { print PROB "\n        "; }
                }    
            }                          
        }

        # Added on July 2nd, 2008. We want to minimize sum d
        if ($train_file ne "none" && ($max_margin == 2 || ($max_margin == 1 && $dimcg_exponent == 1)))
        {            
            $a_param == 2 || die ("a_param 0 highly outdated, NOT IMPLEMENTED any more");
            $unique_index = 0;

            if ($one_delta_permol == 1 && $max_margin == 2)
            {                
                for (my $perm=0; $perm < $num_molecules; $perm++)
                {
                    my $likelihood_normalizer;
                    # Starting Apr 17, 2008, we only use lambda to weigh the thermodynamic set, because now we also include regularizer
                    if ($normalize_all_three_sums)
                    {
                        $likelihood_normalizer = $scaler/$num_molecules;
                    }
                    else
                    {
                        $likelihood_normalizer = $scaler;
                    }
                    if ($lambda == 0 && $unique_index == 0)     { print PROB "    "; }
                    elsif ($exists_delta[$unique_index])        { print PROB " + ";  }
                    if ($exists_delta[$unique_index])
                    {
                        printf PROB (" %e d%d", $likelihood_normalizer, $unique_index);
                    }
                    if ($unique_index != 0 && $unique_index % 10 == 0)
                    { print PROB "\n        "; }
                    $unique_index++;
                }
            }
            else
            {            
                for (my $perm=0; $perm < $num_molecules; $perm++)   # TODO: Not sure if this shouldn't actually be $num_deltas
                {                
                    my $likelihood_normalizer;
    
                    if ($num_constraints_permol[$perm] != 0)
                    {
                        if ($normalize_all_three_sums)
                        {
                            $likelihood_normalizer = $scaler/($num_constraints_permol[$perm]*$num_molecules);
                        }
                        else
                        {
                            $likelihood_normalizer = $scaler/$num_constraints_permol[$perm];
                        }
                    }
                    for (my $j=0; $j < $num_constraints_permol[$perm]; $j++)
                    {
                        # now write the optimization function                    
                        if ($lambda == 0 && $unique_index == 0)     { print PROB "    "; }
                        else                                        { print PROB " + ";  }
                        printf PROB (" %e d%d", $likelihood_normalizer, $unique_index);
                        if ($unique_index != 0 && $unique_index % 10 == 0)
                        { print PROB "\n        "; }
                        $unique_index++;
                        #print "UNIQUE INDEX: $unique_index\n";
                    }
                }
            }
        }

        # done writing the optimization function                    
                
            
        print PROB "\n\nSUBJECT TO\n";        

        # added this on Apr 9, 2008, and modified to add it as constraint on June 5, 2008
        if ($add_regularizer == 1 || $add_regularizer == 2)
        {
            # first get the mean values
            my @means = ();
            my $mline;

            # Apr 17, 2008: Added a regularizer factor
            # June 5, 2008: If it is constraint, I don't need this factor
#             my $factor;
#             if ($normalize_all_three_sums)
#             {
#                 $factor = $gaussian_regularizer_precision * $scaler / $num_params;
#             }
#             else
#             {
#                 $factor = $gaussian_regularizer_precision * $scaler;
#             }
            #my $factor = 1;
            my $coefficient;
            $there_are_non_zero_means = 0;
            if ($regularizer_mean_file eq "0")      # this can be 0, meaning all the means are 0
            {
                for (my $i=0; $i < $num_params; $i++)
                {
                     push (@means, 0);
                }
            }
            else
            {
                open (MEANS, "$directory/$regularizer_mean_file") || die ("Cannot open $directory/$regularizer_mean_file");
                while ($mline = <MEANS>)
                {
                    chomp ($mline);
                    if ($mline != 0)    {   $there_are_non_zero_means = 1;  }
                    push (@means, $mline);
                }
                close (MEANS);
            }
            # sanity check
            if (@means."" != $num_params)
            {
                die ("The regularizer mean file does not have $num_params values!\n");
            }


            # This was when I was adding the regularizer in the objective function
            # Now I'm adding it as constraint    
            # I'm keeping the comments in, in case we want to revert to that
#             if ($train_file eq "none" && $lambda == 0)
#             {
#                 printf PROB ("       [ %e x0^2 ", $factor);
#                 #printf PROB ("       [ %e x0^2 %+e x0 ", $factor, 2*$factor*$means[0]);
#             }
#             else
#             {
#                 printf PROB ("       + %e x0^2 ", $factor);
#                 #printf PROB ("       + %e x0^2 %+e x0 ", $factor, 2*$factor*$means[0]);
#             }
#             for (my $i=1; $i < $num_params; $i++)
#             {
#                 printf PROB (" + %e x%d^2 ", $factor, $i);
#                 #printf PROB (" + %e x%d^2 %+e x%d ", $factor, $i, 2*$factor*$means[$i], $i);
#                 if ($i % 10 == 0)
#                 { print PROB "\n        "; }
#             }             
#             
#             print PROB " ] / 2\n        ";
#             # the first order terms have to be outside of the /2
#             my $terms = 1;
#             for (my $i=0; $i < $num_params; $i++)
#             {
#                 if ($means[$i] != 0)
#                 {
#                     printf PROB (" %+e x%d ", $factor*$means[$i], $i);
#                     $terms++;
#                     if ($terms % 10 == 0)
#                     { print PROB "\n        "; }
#                 }
#             }


            print PROB "    rc: ";
            my $i;
            if ($add_regularizer == 2)   {   print PROB " [ ";   }
            for ($i=0; $i < $num_params; $i++)
            {
                if ($add_regularizer == 1)        # lasso
                {   print PROB "y$i";   }
                elsif ($add_regularizer == 2)
                {   print PROB "x$i^2";    }
                if ($i < $num_params-1)
                {   print PROB " + ";  }
                if ($i != 0 && $i % 10 == 0)
                { print PROB "\n        "; }
            }
            if ($add_regularizer == 2)   {   print PROB " ]  ";   }
            # if the mean is not zero, add the terms
            if ($there_are_non_zero_means)
            {
                my $term = "";
                if ($add_regularizer == 2)
                {
                    for ($i=0; $i < $num_params; $i++)
                    {
                        if ($means[$i] != 0)
                        {
                            my $factor = -2*$means[$i];
                            if ($factor < 0)    {   $term .= " $factor x$i";    }
                            else                {   $term .= " + $factor x$i";    }
                        }
                        if ($i != 0 && $i % 10 == 0)
                        { $term .= "\n        "; }
                    }
                }
                print PROB "$term";
            }
            # the right hand side
            # added the normalization on June 27, 2008
            my $rhs = $regularizer_bound_eta*($num_params-$num_zero_params);
            for ($i=0; $i < $num_params; $i++)
            {
                if ($add_regularizer == 1)      {   $rhs += $means[$i];     }
                elsif ($add_regularizer == 2)   {   $rhs -= $means[$i]*$means[$i];  }
            }
            # normalize this inequality, so that eta doesn't depend on the number of parameters in the model
            # the given eta should be small (around 1 or so)
            
            #$rhs *= ($num_params-$num_zero_params);
            print PROB " <= $rhs\n\n";
        
            if ($add_regularizer == 1)
            {
                for ($i=0; $i < $num_params; $i++)
                {
                    print PROB "    lr1_$i: y$i >= 0\n";
                    print PROB "    lr2_$i: y$i - x$i >= 0\n";
                    print PROB "    lr3_$i: y$i + x$i >= 0\n";
                }   
            }
            # done writting the regularizer constraint        
        }

        # add the constraints for the case DIM-CG with exponent 2
                        
        if ($train_file ne "none" && $max_margin == 1 && $dimcg_exponent == 2)
        {
            my $i = 0;
            for (my $perm=0; $perm < $num_molecules; $perm++)
            {
                for (my $j=0; $j < $num_constraints_permol[$perm]; $j++)
                {
                    print PROB "    lf1_$i: f$i >= 0\n";
                    print PROB "    lf2_$i: f$i - d$i >= 0\n";
                    print PROB "    lf3_$i: f$i + d$i >= 0\n";
                    $i++;
                }
            }
        }
        close (PROB);
        if ($train_file ne "none")
        {
            if ($max_margin == 1 || ($max_margin == 2 && $use_equalities))
            {
                # replace the "<" by "="
                open (CONSTR, $struct_constr_file) || die ("Cannot open file $struct_constr_file");
                open (PROB, ">>$problem_file") || die ("Cannot open file $problem_file");
                while ($line = <CONSTR>)
                {
                    $line =~ s/</=/;
                    print PROB $line;
                }
                close (PROB);
                close (CONSTR);
            }        
            else    # if ($max_margin == 0) or 2
            {
                system ("cat $struct_constr_file >> $problem_file");
            }
        }
        
        # add the thermodynamic constraints
        if ($lambda > 0)
        {
            system ("cat $thermo_constraints_file >> $problem_file");
            # add the regularizer on the parameters not met in the thermo constraints
#             if ($regularizer_beta > 0)
#             {
#                 open (PROB, ">>$problem_file") || die ("Cannot open $problem_file");
#                 print PROB "    \\* Adding p+-x inequalities *\\\n";
#                 for (my $i = 0; $i < $num_params; $i++)
#                 {
#                     
#                     if (! $params_met[$i])
#                     {
#                         print PROB "    pc${i}_m: p$i - x$i >= 0\n";
#                         print PROB "    pc${i}_p: p$i + x$i >= 0\n";
#                     }    
#                 }
#                 close (PROB);
#             }
        }    

        # also add the additional constraints
        if ($additional_constraints_file ne "none")
        {
            system ("cat $directory/$additional_constraints_file >> $problem_file");
        }


        ##################################################
        # add the bounds
        ##################################################
        
        open (PROB, ">>$problem_file") || die ("Cannot open $problem_file");
        print PROB "\nBOUNDS\n";

        # If some parameters are more restricted than this, they should be given in the file with additional constraints
        if ($amount_lub eq "inf")        
        {
            # the dangling ends should be negative or 0
            for (my $i=0; $i < $num_params; $i++)
            {
                print PROB "    -infinity <= x$i <= +infinity\n";
            }
        }
        else
        {
            if ($bounds_from_T_or_P eq "P")
            {
                open (PREV, $params_file) || die ("Cannot open file $params_file");
                my $line_prev;
                # not implemented if amount_lub is percent, e.g. 0.8%
                my $margin = $amount_lub;
                $margin =~ s/K//;
                my $left;
                my $right;
                my $i=0;
                while ($line_prev = <PREV>)
                {
                    chomp($line_prev);
                    $left = $line_prev - $margin;
                    $right = $line_prev + $margin;
                    print PROB "    $left <= x$i <= $right\n";                    
                    $i++;
                }
                close (PREV);
            }
            else
            {
                # TODO: These are read earlier too, I think, I can just use those values
                open (LB, "$newdir/params_lb_${amount_lub}.txt") || die ("Cannot open LB");
                open (UB, "$newdir/params_ub_${amount_lub}.txt") || die ("Cannot open UB");
                my $i=0;
                while ($line_lb = <LB>)
                {
                    chomp ($line_lb);
                    $line_ub = <UB>;
                    chomp ($line_ub);          
                    # for now use bounds dependent on Turner params, because the problem is still way under-constraint              
                    #if ($thermo eq "thermo-yes")
                    #{
                    #    print PROB "    -10 <= x$i <= 10\n";
                    #}
                    #else
                    {
                        # make sure the dangling ends are negative
#                         if ($i >= $first_d3 && $i <= $last_d5)
#                         {
#                             $line_ub = $line_ub < 0? $line_ub:0;
#                         }
                        # fix some of the parameters, according to the fixparams_file, if given
                        if ($fixparams_file eq "none")
                        {
                            print PROB "    $line_lb <= x$i <= $line_ub\n";
                        }
                        else
                        {
                            if ($fixparams[$i] eq "variable")
                            {
                                print PROB "    $line_lb <= x$i <= $line_ub\n";
                            }
                            else
                            {
                                print PROB "    x$i = $fixparams[$i]\n";
                            }
                        }
                    }    
                    $i++;
                }
                close (LB);
                close (UB);
            }    
        }    
        # we don't need to explicitly add bounds for d variables, they have to be >= 0, which is default in CPLEX
        if ($lambda > 0)
        {
            # add the bounds on v: they can be positive or negative
            for (my $i=0; $i < $num_thermo_constr; $i++)
            {
                print PROB "    -infinity <= v$i <= +infinity\n";
            }
        }

        # Added on July 2nd, when I added the max margin
        if ($max_margin == 1)
        {
            for (my $i=0; $i < $num_struct_constr; $i++)
            {
                print PROB "    -infinity <= d$i <= +infinity\n";
            }
        }

        elsif ($max_margin == 2 && $one_delta_permol)
        {
            for (my $i=0; $i < $num_molecules; $i++)
            {
                if ($exists_delta[$i])
                {
                    print PROB "    -infinity <= d$i <= +infinity\n";
                }
            }
        }

        # just to check Kevin's perceptron idea: make the deltas (d) unbounded
#         if ($normalize eq "n1")
#         {
#             for (my $i=0; $i < $num_struct_constr; $i++)
#             {
#                 print PROB "    -infinity <= d$i <= +infinity\n";
#             }
#         }
        print PROB "\nEND\n";
        close (PROB);

        # now the input to CPLEX is created
        
        # now create the CPLEX script
        $command = "echo \"read $problem_file_nopath\" > $newdir/prog";
        print OPT "$indent$command\n";
        system($command);        
        
        # now read the basis file, if it exists. Gave up using the basis file.
        #my $isbasis = `ls $prev_basis_file`;
        #if ($isbasis > -1)    # exists
        #{
        #    $command = "echo \"read $prev_basis_file\" >> $newdir/prog";
        #    print "$indent$command\n";
        #    system($command);        
        #}    
                
        # try to use parallel cplex, but:
        # it looks like we don't have licence for parallel cplex
        #$command = "echo \"set threads 2\" >> $newdir/prog";
        #system($command);
        #$command = "echo \"baropt\" >> $newdir/prog";
        

        # decide which optimizer to use
        $command = "echo \"$optimizer\" >> $newdir/prog";
        
        print OPT "$indent$command\n";
        system($command);        
        $basis_file = $problem_file;
        $basis_file =~ s/problem/basis/;
        $basis_file =~ s/\.lp/\.bas/;
        
        $command = "echo \"write $answer_file_nopath\" >> $newdir/prog";
        print OPT "$indent$command\n";
        system($command);            
        # write the basis solution, so that next time the optimization is faster
        #$command = "echo \"write $basis_file\" >> $newdir/prog";
        #print "$indent$command\n";
        #system($command);

        ##################################################
        # Now start running CPLEX
        ##################################################
             
        if ($use_sge_instead_of_mpi)
        {
            # try to run cplex directly from the directory $newdir
            my $sge_cplex_nopath = "sge_cplex.sh";
            my $sge_cplex = "$newdir/$sge_cplex_nopath";
            print OPT "${indent}Creating the sge cplex script $sge_cplex\n";
            open (SGE, ">$sge_cplex") || die ("Cannot open $sge_cplex");
            print SGE "#!/bin/sh\n\n";

            # for safety, make sure there are cplex licences available
            # CHecking cplexusers doesn't seem to be good enough, so be more aggressive
            
            print SGE "while [ ! -e $answer_file_nopath ]\n";
            print SGE "do\n";
            print SGE "     /cs/local/lib/pkg/cplex-10.1.1/bin/cplex < prog | grep -v \"does not exist\" >> $output_verbose 2>&1\n";
            print SGE "     sleep 1\n";
            print SGE "done\n\n";
#                 print SGE "NO=`/cs/local/lib/pkg/cplex-10.1.1/bin/cplexusers | grep \"available tokens\" | awk '{print \$3;}'`\n";
#                 #print SGE "echo \$NO\n";
#                 print SGE "while [ \$NO -eq 0 ]\n";
#                 print SGE "do\n";
#                 #print SGE "    echo \$NO\n";
#                 print SGE "    sleep 1\n";
#                 print SGE "    NO=`cplexusers | grep \"available tokens\" | awk '{print \$3;}'`\n";
#                 print SGE "done\n";
#                 print SGE "/cs/local/lib/pkg/cplex-10.1.1/bin/cplex < $newdir/prog | grep -v \"does not exist\" >> $options_file\n\n";
            close(SGE);
        
            if ($cluster_for_cplex eq "arrow")
            {
                $command = "nohup ssh samos \"source /cs/beta/lib/pkg/sge-6.0u7_1/default/common/settings.csh; cd $newdir; perl start_sge_job.pl 'qsub -cwd -S /bin/sh -o /dev/null -e /dev/null -P kpm $sge_cplex_nopath'\"";
                # -l cplex=1
                #
                print OPT "${indent}Sending cplex job to arrow cluster: $command\n";
            
                # I have to close it, otherwise what I write with print OPT overwrites what I write with >> $options_file
                close (OPT);
                my $commout = `$command`;
                open (OPT, ">>$options_file") || die ("Cannot open $options_file");
                print OPT "${indent}$commout";
            }
            elsif ($cluster_for_cplex eq "icics")   # actually
            {
                $command = "cd $newdir; perl start_sge_job.pl 'qsub -p -1 -cwd -S /bin/sh -o /dev/null -e /dev/null -q icics.q $sge_cplex_nopath'";
                #$command = "cd $newdir; perl start_sge_job.pl 'qsub -cwd -S /bin/sh -o /dev/null -e /dev/null -q icics.q $sge_cplex_nopath'";
                #
                print OPT "${indent}Sending cplex job to icics cluster: $command\n";
            
                # I have to close it, otherwise what I write with print OPT overwrites what I write with >> $options_file
                close (OPT);
                my $commout = `$command`;
                open (OPT, ">>$options_file") || die ("Cannot open $options_file");
                print OPT "${indent}$commout";
            }
        }                      
        else    # don't use sge, just run on the current computer
        {              
            # this is what I had before, before invoking sge for cplex                      
            $command = "cd $newdir; cplex < prog | grep -v \"does not exist\" >> $output_verbose 2>&1";
            print OPT "$indent$command\n";
            # I have to close it, otherwise what I write with print OPT overwrites what I write with >> $options_file
            close (OPT);
            system("$command");
            open (OPT, ">>$options_file") || die ("Cannot open $options_file");
        }               
        
        
        # now parse the answer file and write the results into results
        print OPT "${indent}Parsing CPLEX results ... writting $new_params_file\n";
        #system ("cp $params_file $new_params_file");
        parse_CPLEX_answer ($answer_file, $new_params_file, $opt_problem);
    }    
    elsif ($solver eq "MATLAB")
    # I'm pretty sure this DOES NOT work! It's waaay outdated
    {
        #$command = "mv matrix_A.txt results/matrix-A_num-${num_seqs}_sub-${s}.txt";
        #print "$indent$command\n";
        #system($command);
        #$command = "mv matrix_B.txt results/matrix-B_num-${num_seqs}_sub-${s}.txt";
        #print "$indent$command\n";
        #system($command);
        $command = "echo \"clear all\" > prog.m";
        print OPT "$indent$command\n";
        system($command);
        #if ($optimized eq "opt=trusted")
        #{
        #    $command = "echo \"optimize_trusted(${num_seqs},$s)\" >> prog.m";
        #}    
        #elsif ($optimized eq "opt=noconstr")
        #{
        #    $command = "echo \"optimize_noconstr(${num_seqs},$s)\" >> prog.m";
        #}      
        #if ($optimized eq "opt=epsilon")
        #{
            $command = "echo \"optimize_epsilon(${num_train},$s)\" >> prog.m";
        #}  
        #else
        #{
        #    $command = "echo \"optimize(${num_seqs},$s)\" >> prog.m";
        #}      
        print OPT "$indent$command\n";
        system($command);
        $command = "matlab -nodisplay -nojvm < prog.m";
        print OPT "$indent$command\n";
        system($command);
        $command = "echo \"clear all\" > prog.m";
        print OPT "$indent$command\n";
        system($command);
        $command = "echo \"saveres(${num_train},$s)\" > prog.m";
        print OPT "$indent$command\n";
        system($command);
        $command = "matlab -nodisplay -nojvm < prog.m";
        print OPT "$indent$command\n";
        system($command);
    }

    # Sometimes the number of parameters at this iteration $s are the same as at the previous iteration $tmp.
    # If so, stop.
    # Added on August 9, 2008
    my $diffres = `diff $new_params_file $params_file`;
    chomp ($diffres);
    if ($diffres eq "")     # they are the same
    {
        system ("echo \"===============\" >> $results_file");
        system ("echo \"STOP! Parameters at iterations $s are the same as parameters at iteration $tmp\" >> $results_file");
        system ("rm -f $datafile");
        system ("rm -rf $datadir");
        last;        
    }
    
    system ("echo \"===============\" >> $results_file");                          
    system ("echo \"Num subopt: $s\" >> $results_file");

    ##################################################
    # Now analyse the results obtained with the new set of parameters,
    #   and predict new structures for the next iteration
    ##################################################

    # first check the datafile has num_train tuples
    

    if ($train_file ne "none")
    {
        my $num_tuples;
        $num_tuples = `cat $datafile | grep "^>" | wc -l`;
        chomp ($num_tuples);
        if ($iteration == 1)
        {
            if ($num_tuples < $num_train)
            {
                print OPT "ERROR! File $datafile should have at least $num_train tuples and it has $num_tuples\n";
                print "ERROR! File $datafile should have at least $num_train tuples and it has $num_tuples\n";
                exit;
            }
        }
        else
        {
            if ($num_tuples != $num_train)
            {
                print OPT "ERROR! File $datafile should have $num_train tuples and it has $num_tuples\n";
                print "ERROR! File $datafile should have $num_train tuples and it has $num_tuples\n";
                exit;
            }
        }    
        
    
        # see if we want to use MPI or just SGE
        if ($use_sge_instead_of_mpi)    # use sge
        {
            # first create the new datadir directory
            system ("mkdir $new_datadir");
            system ("mkdir $datadir/tmp");  # we might need tmp for hotknots
            # prepare the sge qsub command to run the prediction and analysis on the cluster        
            print OPT "${indent}Starting an SGE array job on cluster $cluster_for_prediction for prediction and analysis on the training set, with the new parameters\n";
            
            
            my $delta_file;
            if ($add_constraint_if_violation)
            {
                print OPT "Creating the delta files in directory $newdir/setdir$tmp\n";
                # create many delta files instead of just one
                #get_deltas_from_CPLEX_answer ($answer_file, "$newdir/$delta_file");
                # one delta file corresponding to each file in setdir$tmp
                create_delta_files_from_CPLEX_answer ($answer_file, "$newdir/setdir$tmp");
            }            
            if ($cluster_for_prediction eq "icics")
            {
                $command = "cd $newdir; perl start_sge_job.pl 'qsub -p -1 -cwd -S /bin/sh -o /dev/null -e /dev/null -t 1000-". (1000+$num_split_files-1)." $sge_predict_and_analyse params_sub-${s}.txt setdir$tmp setdir$s '";
                #$command = "cd $newdir; perl start_sge_job.pl 'qsub -cwd -S /bin/sh -o /dev/null -e /dev/null -t 1000-". (1000+$num_split_files-1)." $sge_predict_and_analyse params_sub-${s}.txt setdir$tmp setdir$s '";
            }
            elsif ($cluster_for_prediction eq "arrow")
            {
                $command = "nohup ssh samos \"source /cs/beta/lib/pkg/sge-6.0u7_1/default/common/settings.csh; cd $newdir; perl start_sge_job.pl 'qsub -cwd -S /bin/sh -o /dev/null -e /dev/null -p -3 -P kpm -t 1000-". (1000+$num_split_files-1)." $sge_predict_and_analyse params_sub-${s}.txt setdir$tmp setdir$s '\"";
            }          
            # if there is any error above, it's written in $results_file. So check there for ERROR
            check_error($results_file);            
    
            print OPT "$indent$command\n";
            # run it via the wrapper which exits only when the jobs are done
            # added the wrapper above
            system ($command);
            # once this finished, I have to collect and average the accuracy results, and add it to the results file
            $command = "perl average_analysed_results.pl $new_datadir $s \"Training set\" >> $results_file";
            print OPT "$indent$command\n";
            system ($command);

            check_error($results_file);
            
            # create the datafile, it's needed at later iterations                          
            # cat doesn't work when we have too many files. Too long argument.
            # system ("cat $new_datadir/file*.txt > $new_datafile");
            # so I created my own cat
            system ("perl cat_all_files.pl $new_datadir/file $new_datafile");

            #TODO: the perturbation thingie

            # Now test the performance on the validation set, if it exists
            # Added on June 3, 2008
            if ($validation_file ne "none")
            {                                                               
                # there must be a directory with files called $newdir/valdir0

                # first create the new datadir directory
                my $new_valdir = "$newdir/valdir";
                system ("mkdir $new_valdir");
                # prepare the sge qsub command to run the prediction and analysis on the cluster
                print OPT "${indent}Starting an SGE array job on cluster $cluster_for_prediction for prediction and analysis on the validation set, with the new parameters\n";

                if ($cluster_for_prediction eq "icics")
                {
                    $command = "cd $newdir; perl start_sge_job.pl 'qsub -p -1 -cwd -S /bin/sh -o /dev/null -e /dev/null -t 1000-". (1000+$num_split_files_validation-1)." $sge_validate params_sub-${s}.txt valdir0 valdir '";
                    #$command = "cd $newdir; perl start_sge_job.pl 'qsub -cwd -S /bin/sh -o /dev/null -e /dev/null -t 1000-". (1000+$num_split_files_validation-1)." $sge_validate params_sub-${s}.txt valdir0 valdir '";
                }
                elsif ($cluster_for_prediction eq "arrow")
                {
                    $command = "nohup ssh samos \"source /cs/beta/lib/pkg/sge-6.0u7_1/default/common/settings.csh; cd $newdir; perl start_sge_job.pl 'qsub -cwd -S /bin/sh -o /dev/null -e /dev/null -p -1 -P kpm -t 1000-". (1000+$num_split_files_validation-1)." $sge_validate params_sub-${s}.txt valdir0 valdir '\"";
                }
                # if there is any error above, it's written in $results_file. So check there for ERROR
                check_error($results_file);
        
                print OPT "$indent$command\n";
                # run it via the wrapper which exits only when the jobs are done
                # added the wrapper above
                system ($command);
                # once this finished, I have to collect and average the accuracy results, and add it to the results file
                $command = "perl average_analysed_results.pl $new_valdir $s \"Validation set\" >> $results_file";
                print OPT "$indent$command\n";
                system ($command);        
                check_error($results_file);
                # don't forget to remove the new valdir                    

                system ("rm -rf $new_valdir");
            }
            
        }
        else    # use MPI
        {
            # check if num_train is less than the number of processors
            my $num_processors = `cat hostfile | wc -l`;
            chomp ($num_processors);
            my $runproc;
            if ($num_processors > $num_train+1)
            {
                $runproc = "c0-$num_train";
            }
            else
            {
                $runproc= "C";
                #$runproc= "c0-1";
            }
                    
            if (! $use_several_computers)
            {
                $runproc = "c0-1";
            }
            if ($validation_file eq "none")
            {
                $command = "mpirun $runproc ./predict_and_analyse_results_mpi $new_params_file $datafile $new_datafile $num_train $counts_file $count_threshold $simple_model $replace_instead_of_add none >> $results_file ";
            }
            else
            {    
                $command = "mpirun $runproc ./predict_and_analyse_results_mpi $new_params_file $datafile $new_datafile $num_train $counts_file $count_threshold $simple_model $replace_instead_of_add $newdir/validation_set.txt >> $results_file ";
            }
            #$replace_instead_of_add
            #$command = "./analyse_results $new_params_file $datafile $new_datafile $num_train >> $results_file ";
            #$command = "./analyse_results $newdir $new_params_file $train_file $num_train $test_file $num_test $s >> $results_file ";
            print OPT "$indent$command\n";
            system("nice $command");
        
            # check if the mpirun ran ok
            mpirun_validate();
            
            my $to_stop = `tail -6 $results_file | grep STOP`;
        
            if ($to_stop =~ /STOP/)
            {
                if ($do_perturbation)
                {
                    # try to perturb the last parameters, so that we can search further
                    # TODO: this perturb somehow contradicts the idea of "no bounds in the code", and so it needs "simple_model". TO FIX.
                    system("perl perturb_parameters.pl $new_params_file tmppar.txt $simple_model");
                    system("mv tmppar.txt $new_params_file");
        
                    # rerun mpi
                    if ($validation_file eq "none")
                    {    
                        $command = "mpirun $runproc ./predict_and_analyse_results_mpi $new_params_file $datafile $new_datafile $num_train $counts_file $count_threshold $simple_model $replace_instead_of_add none >> $results_file ";
                    }
                    else
                    {    
                        $command = "mpirun $runproc ./predict_and_analyse_results_mpi $new_params_file $datafile $new_datafile $num_train $counts_file $count_threshold $simple_model $replace_instead_of_add $newdir/validation_set.txt >> $results_file ";
                    }
                    
                    #$replace_instead_of_add
                    #$command = "./analyse_results $new_params_file $datafile $new_datafile $num_train >> $results_file ";
                    #$command = "./analyse_results $newdir $new_params_file $train_file $num_train $test_file $num_test $s >> $results_file ";
                    print OPT "$indent$command\n";
                    system("nice $command");
                    
                    # check if the mpirun ran ok
                    mpirun_validate();
                }
                else    
                {
                    print OPT "STOPPED after $s iterations because no more structures could be generated\n";
                    exit;
                }
            }
        }
        # only now remove the previous data file, if mpirun ran ok
        # TODO: remove the $datadir
    
        system ("rm -f $datafile");
        system ("rm -rf $datadir");
    }   # end if ($train_file ne "none")

    # exit if no training file
    if ($train_file eq "none")
    { last; }
}   # end for (number of iterations)



# open results file, and automatically get the best parameter set, and test it on the test data

if ($test_file ne "none")           {   $test_file = "$directory/$test_file";   }

$command = "perl pick_best_training.pl $newdir $test_file $use_sge_instead_of_mpi $min_length_sum";
print OPT "$command\n";
system ($command);

close (OPT);

####################################################


sub check_error
{
    my ($file) = @_;
    
    #CHeck if there are ERROR words in $results_file
    my $error_result = `grep ERROR $file`;
    if ($error_result ne "")
    {
        # Some error happened.
        print "ERROR written in $file!\n";
        print $error_result;
        exit;
    }

}

####################################################
sub parse_IPOPT_answer
{
    my ($answer_file, $params_file) = @_;
    my $line;
    open (ANS, $answer_file) || die ("Cannot open $answer_file");
    open (RES, ">$params_file") || die ("Cannot open file $params_file");            

    while ($line = <ANS>)
    {
        # Check to see if it found the optimal solution, otherwise don't waste the CPUs
        if ($line =~ /EXIT/)
        {
            if ($line =~ /Optimal Solution Found/)  # great
            {}
            else
            { die ("ABORT, IPOPT didn't find optimal solution, message is:\n$line"); }
        }
        if ($line =~ /Solution of the primal variables, x/) {   last;   }
    }
    for (my $i=0; $i < $num_params; $i++)
    {
        $line = <ANS>;
        if ($line =~ /x\[$i\] = (\S+)/)
        {
            printf RES ("%.2f\n", $1);
        }
        else
        {
            print "ERROR! Line doesn't contain x[$i]\n$line\n";
            exit;
        }
    }
    close (ANS);
    close (RES);
}

####################################################

sub parse_CPLEX_answer
{
    my ($answer_file, $params_file, $opt_problem) = @_;
    my $line;
    #print "PARAM file: $params_file\n";
    
    # first get Turner's parameters, so that we have these when no parameters could be figured out
    my @turner_params = ();
    open (TUR, "$directory/$initial_parameters") || die ("Cannot open $directory/$initial_parameters");
    while ($line = <TUR>)
    {
        chomp ($line);
        push (@turner_params, $line);        
        # trying with 0 instead of Turner gives rubbish 
        #    before I was replacing everything with is 0 with Turner's param, in params.cpp
        #    but that wasn't correct, because some parameters can be estimated as 0 by cplex.
        #    better to put Turner params here for those params which were not met in the optimization problem.
        #push (@turner_params, 0.00);
    }
    close (TUR);
    
    # now parse the Answer file
    open (ANS, "$answer_file") || die ("Cannot open file $answer_file");    
    if ($opt_problem eq "lp")    # so txt format
    {
        while ($line = <ANS>)
        {
            if ($line =~ /STATUS/)
            {
                if ($line !~ /OPTIMAL SOLN/)
                {
                    print OPT "ABORT: Solution is not optimal at $answer_file\n$line\n";
                    print "ABORT: Solution is not optimal at $answer_file\n$line\n";
                    exit;
                }
            }
            elsif ($line =~ /x(\d+)\s+\w+\s+(\-*\d+\.*\d*)/)
            {
                $turner_params[$1] = $2;
                #print RES "$1 $2\n";
            }    
        }
    }
    elsif ($opt_problem eq "qp")
    # with CPLEX version 10.1.1, qp is saved in sol format
    # with version 9.1, it also worked in .vec format
    {
        # can't check status
        if ($answer_file =~ /\.vec/)    # vec format
        {
            while ($line = <ANS>)
            {
                if ($line =~ /c\s+x(\d+)\s+(\-*\d\.\d+e[\-+]\d\d)/)
                {
                    $turner_params[$1] = $2;
                }
            }
        }
        elsif ($answer_file =~ /\.sol/)   # SOL format
        {
            # a line of interest looks like this:
            # <variable name="x7" index="1006" value="-2.62290759284369" reducedCost="-3.12259412378588e-11"/>
            while ($line = <ANS>)
            {
                if ($line =~ /variable name=\"x(\d+)\".*value=\"(.+)\"/)
                {
                    $turner_params[$1] = $2;
                }
            }
        }
        else
        {
            print OPT "ERROR: answer_file should be VEC or SOL\n";
            print "ERROR: answer_file should be VEC or SOL\n";
            exit;
        }
    }    
    else
    {
        print OPT "ERROR: opt_problem must be lp or qp, and is: |$opt_problem|\n";
        print "ERROR: opt_problem must be lp or qp, and is: |$opt_problem|\n";
        exit;
    }
    close (ANS);
    
    open (RES, ">$params_file") || die ("Cannot open file $params_file");    
    foreach my $par (@turner_params)
    {
        printf RES ("%.5f\n", $par);
    }    
    close (RES);    
}

####################################

sub create_delta_files_from_CPLEX_answer
{
    my ($answer_file, $strdir) = @_;
    my $line;

    # first get the deltas from the CPLEX answer file
    my @deltas = ();
    #print "PARAM file: $params_file\n";

    for (my $i=0; $i < $num_molecules; $i++)
    {
        $deltas[$i] = 0;
    }
    
    # now parse the Answer file
    open (ANS, "$answer_file") || die ("Cannot open file $answer_file");    
    $opt_problem eq "qp" || die ("lp format not supported any more!! ABORT");
    $answer_file =~ /\.sol/ || die ("The answer file $answer_file must be in .sol");
    # a line of interest looks like this:
    # <variable name="d0" index="0" value="18.5919479365155" reducedCost="6.64693206617155e-11"/>
    while ($line = <ANS>)
    {
        if ($line =~ /variable name=\"d(\d+)\".*value=\"(.+)\"/)
        {
            $deltas[$1] = $2;
        }
    }
    close (ANS);

    # now, for each $strdir/res(\d+).txt, create one delta file
    my $last_file = `ls $strdir/file*.txt | wc -l`;
    chomp ($last_file);
    $last_file += 999;
    my $current_index = 0;
    for (my $i=1000; $i <= $last_file; $i++)
    {
        my $num_rows = `cat $strdir/file$i.txt | grep "^>" | wc -l`;
        chomp ($num_rows);
        #print "NUm rows for file $i is $num_rows\n";
        # that's how many rows I have to write in delta$i.txt
        open (DEL, ">$strdir/delta$i.txt") || die ("Cannot open file $strdir/delta$i.txt");
        for (my $j = $current_index; $j < $current_index + $num_rows; $j++)
        {
            print DEL "$deltas[$j]\n";            
        }
        $current_index += $num_rows;
        close (DEL);
    }
}

####################################


sub get_deltas_from_CPLEX_answer
{
    my ($answer_file, $output_deltas_file) = @_;
    my $line;
    my @deltas = ();
    #print "PARAM file: $params_file\n";

    for (my $i=0; $i < $num_molecules; $i++)
    {
        $deltas[$i] = 0;
    }
    
    # now parse the Answer file
    open (ANS, "$answer_file") || die ("Cannot open file $answer_file");    
    $opt_problem eq "qp" || die ("lp format not supported any more!! ABORT");
    $answer_file =~ /\.sol/ || die ("The answer file $answer_file must be in .sol");
    # a line of interest looks like this:
    # <variable name="d0" index="0" value="18.5919479365155" reducedCost="6.64693206617155e-11"/>
    while ($line = <ANS>)
    {
        if ($line =~ /variable name=\"d(\d+)\".*value=\"(.+)\"/)
        {
            $deltas[$1] = $2;
        }
    }
    close (ANS);

    open (DEL, ">$output_deltas_file") || die ("Cannot open $output_deltas_file");
    foreach my $delta (@deltas)
    {
        print DEL ($delta+$epsilon_for_violation)."\n";
    }
    close (DEL);
}

####################################

sub get_num_points
{
    my ($train) = @_;
    my $num = `grep "^>" $train | wc -l`;
    chomp ($num);
    return $num;
}

####################################

sub mpirun_validate
{
    # check if everything was ok with mpirun
    if ($? == -1) {
        print OPT "ERROR in mpirun: failed to execute: $!\n";
        print "ERROR in mpirun: failed to execute: $!\n";
        exit;
    }
    elsif ($? & 127) {
        printf OPT "ERROR in mpirun: child died with signal %d, %s coredump\n",
            ($? & 127),  ($? & 128) ? 'with' : 'without';    
        printf "ERROR in mpirun: child died with signal %d, %s coredump\n",
            ($? & 127),  ($? & 128) ? 'with' : 'without';
        exit;
    }
    else {
        # everything is good
        #printf OPT "child exited with value %d\n", $? >> 8;
        #printf "child exited with value %d\n", $? >> 8;
    }
}

####################################

sub get_derivative
# used for IPOPT, to get the partial derivative of $function, wrt the variable x[$variable]
# return a text with the partial derivative, or "" in case the derivative is 0
# example of function
#  2 * x[7] * x[369]  + 2 * x[8] * x[369]  + 2 * x[12] * x[369]  + (-2) * x[6]  + (-3) * x[7]  + (-1) * x[8]  + (-1) * x[12]  + (-1) * x[57]  + (-2) * x[117]  + (-1) * x[118]  + (-1) * x[308]  + (-1) * x[309]  + (-1) * x[311]  + (-1) * x[320]  + 1 * x[363]  + 2 * x[366]  + 10 * x[367]
# This function never has x[i] * x[i] (i.e. square of the same variable)
{
    my ($function, $variable) = @_;
    my $derivative = "";
    # first check if $variable is even there
    if ($function =~ /x\[$variable\]/)      # then we have the variable
    {
        # function can be quadratic or just linear
        # first split up the terms
        my @terms = split (/\+/, $function);
        foreach my $term (@terms)
        {
            if ($term =~ /x\[$variable\]/)      # then we have the variable
            {
                # eliminate the variable and the * in front, and that's the derivative
                $term =~ s/\s\*\sx\[$variable\]//;
                # sanity check
                $term !~ /x\[$variable\]/ || die ("ERROR! $term is supposed to be the derivative of x[$variable], but still has it! $term");
                if ($derivative ne "")  {   $derivative .= " + ";   }
                $derivative .= $term;
            }
        }
    }
    else    {   return "";  }
    return $derivative;
}




